__author__ = "N. Vasios"
__email__ = "vasios@g.harvard.edu"

'''
Numerical Analysis of Periodic Laminar and Fibrous media undergoing aJamming Transition
N. Vasios, B. Aktas, Y. Narang, R. Howe, K. Bertoldi

Abaqus python script for the numerical analysis of fiber RVEs in a staggered packing configuration
'''

from part import *
from material import *
from section import *
from assembly import *
from step import *
from interaction import *
from load import *
from mesh import *
from optimization import *
from job import *
from sketch import *
from visualization import *
from connectorBehavior import *
import numpy as np

execfile("utilities.py")

## Naming Parts, Materials, Instances, Jobs etc
ModelName = 'Model-1'


#--------------------------------------------------------
# USER DEFINED PARAMETERS
#

# Model dimensions
L = 1.0 # Fiber width, and depth

# Gap between fibers and tolerance for node-based selection
gap = 1e-2
tol = 1e-5

# Percentage of fillet radius (0=Square Fibers, 100=Circular Fibers)
fillet_radius_percentage = 0

# Number of finite elements per dimension
numel_dir = 5

# Material Parameters
E,nu = 2000, 0.42  # Data for polypropelene plastic

# Value of pressure to be applied in MPa
p = 0.101

# Number of cpus
ncpus = 8 
SUBMIT = False

yshift = 0

# Boolean to include slave RVEs
include_slaves = False

# Boolean to use quadratic elements
use_quadratic_elements = False

#--------------------------------------------------------

fillet_radius = fillet_radius_percentage*(0.5/100.0)

# Job name and cae file name
JobName = 'Fiber_3D_Staggered_fillet_{0}'.format(int(fillet_radius_percentage))
Savename = JobName + '.cae'

# Global seed size corresponding to the number of elements per dimension chosen
Seedsize = L/numel_dir

# Change directory of abaqus to the directory of the python script
os.chdir(os.path.dirname(os.path.realpath('__file__')))

#--------------------------------------------------------------------------------------

# Use coordinates to represent geometric selections
session.journalOptions.setValues(replayGeometry=COORDINATE, recoverGeometry=COORDINATE)

# Initialize the Model Database
Mdb()

# Get the model of interest
Model = mdb.models['Model-1']

# Use a cartesian coordinate system
Model.rootAssembly.DatumCsysByDefault(CARTESIAN)


#--- Part Definitions

# The coordinates of the points of interest

p1 = (-gap-L,-L-gap,0)

p2 = (-gap,-gap,0)

p3 = (gap,-L-gap,0)

p4 = (L+gap,-gap,0)

p5 = (-0.5*L,gap,0)

p6 = (0.5*L,gap+L,0)

p7 = (2*gap+0.5*L,gap,0)

p8 = (2*gap+1.5*L,gap+L,0)


# Sketch the 2D cross-section of the part
Model.ConstrainedSketch(name='__profile__', sheetSize=L)


# Rectangle 1
Model.sketches['__profile__'].rectangle(
    point1=p1[:-1], 
    point2=p2[:-1])

# Rectangle 2
Model.sketches['__profile__'].rectangle(
    point1=p3[:-1], 
    point2=p4[:-1])

# Add the fillet radius if R>0
if fillet_radius>0:

    # Fillet entities
    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p1[0], p1[1]+0.9*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p1[0]+0.1*L, p1[1]+L), ), 
        nearPoint1=(p1[0],p1[1]+0.9*L), 
        nearPoint2=(p1[0]+0.1*L,p1[1]+L), 
        radius=fillet_radius)

    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p1[0]+L, p1[1]+0.9*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p1[0]+0.9*L, p1[1]+L), ), 
        nearPoint1=(p1[0]+L,p1[1]+0.9*L), 
        nearPoint2=(p1[0]+0.9*L,p1[1]+L), 
        radius=fillet_radius)

    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p3[0], p3[1]+0.9*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p3[0]+0.1*L, p3[1]+L), ), 
        nearPoint1=(p3[0],p3[1]+0.9*L), 
        nearPoint2=(p3[0]+0.1*L,p3[1]+L), 
        radius=fillet_radius)

    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p3[0]+L, p3[1]+0.9*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p3[0]+0.9*L, p3[1]+L), ), 
        nearPoint1=(p3[0]+L,p3[1]+0.9*L), 
        nearPoint2=(p3[0]+0.9*L,p3[1]+L), 
        radius=fillet_radius)

    Model.sketches['__profile__'].delete(objectList=tuple([dim[1] for dim in Model.sketches['__profile__'].dimensions.items()]))

# Make the deformable part
Model.Part(
    dimensionality=THREE_D, 
    name='Intermediate-1', 
    type=DEFORMABLE_BODY)

# Extrude the 2D cross-section by half the depth of the actual part (then mirror)
Model.parts['Intermediate-1'].BaseSolidExtrude(
    depth=0.5*L, 
    sketch=Model.sketches['__profile__'])


# Sketch the 2D cross-section of the part
Model.ConstrainedSketch(name='__profile__', sheetSize=L)


# Rectangle 3
Model.sketches['__profile__'].rectangle(
    point1=p5[:-1], 
    point2=p6[:-1])

# Rectangle 4
Model.sketches['__profile__'].rectangle(
    point1=p7[:-1], 
    point2=p8[:-1])


# Add the fillet radius if R>0
if fillet_radius>0:

    # Fillet entities
    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p5[0], p5[1]+0.1*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p5[0]+0.1*L, p5[1]), ), 
        nearPoint1=(p5[0],p5[1]+0.1*L), 
        nearPoint2=(p5[0]+0.1*L,p5[1]), 
        radius=fillet_radius)

    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p5[0]+L, p5[1]+0.1*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p5[0]+0.9*L, p5[1]), ), 
        nearPoint1=(p5[0]+L,p5[1]+0.1*L), 
        nearPoint2=(p5[0]+0.9*L,p5[1]), 
        radius=fillet_radius)

    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p7[0], p7[1]+0.1*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p7[0]+0.1*L, p7[1]), ), 
        nearPoint1=(p7[0],p7[1]+0.1*L), 
        nearPoint2=(p7[0]+0.1*L,p7[1]), 
        radius=fillet_radius)

    Model.sketches['__profile__'].FilletByRadius(
        curve1=Model.sketches['__profile__'].geometry.findAt((p7[0]+L, p7[1]+0.1*L), ), 
        curve2=Model.sketches['__profile__'].geometry.findAt((p7[0]+0.9*L, p7[1]), ), 
        nearPoint1=(p7[0]+L,p7[1]+0.1*L), 
        nearPoint2=(p7[0]+0.9*L,p7[1]), 
        radius=fillet_radius)

    Model.sketches['__profile__'].delete(objectList=tuple([dim[1] for dim in Model.sketches['__profile__'].dimensions.items()]))

# Make the deformable part
Model.Part(
    dimensionality=THREE_D, 
    name='Intermediate-2', 
    type=DEFORMABLE_BODY)

# Extrude the 2D cross-section by half the depth of the actual part (then mirror)
Model.parts['Intermediate-2'].BaseSolidExtrude(
    depth=0.5*L, 
    sketch=Model.sketches['__profile__'])

# Sketch the 2D cross-section of the part
Model.ConstrainedSketch(name='__profile__', sheetSize=L)


# Make the part that will be used to cut the geometry
Model.sketches['__profile__'].Line(
    point1=(p1[0]+0.5*L,p1[1]+0.5*L),
    point2=(p3[0]+0.5*L,p3[1]+0.5*L))

Model.sketches['__profile__'].Line(
    point1=(p3[0]+0.5*L,p3[1]+0.5*L),
    point2=(p7[0]+0.5*L,p7[1]+0.5*L+yshift))

Model.sketches['__profile__'].Line(
    point1=(p7[0]+0.5*L,p7[1]+0.5*L+yshift),
    point2=(p5[0]+0.5*L,p5[1]+0.5*L+yshift))

Model.sketches['__profile__'].Line(
    point1=(p5[0]+0.5*L,p5[1]+0.5*L+yshift),
    point2=(p1[0]+0.5*L,p1[1]+0.5*L))

Model.sketches['__profile__'].rectangle(
    point1=(-2*L,-2*L), 
    point2=(2*L,2*L))

# Make the deformable part
Model.Part(
    dimensionality=THREE_D, 
    name='Aux', 
    type=DEFORMABLE_BODY)

# Extrude the 2D cross-section by half the depth of the actual part (then mirror)
Model.parts['Aux'].BaseSolidExtrude(
    depth=0.5*L, 
    sketch=Model.sketches['__profile__'])

# Delete the sketch
del Model.sketches['__profile__']


# Instanciate parts
Model.rootAssembly.Instance(
    dependent=ON, 
    name='Aux', 
    part=Model.parts['Aux'])

Model.rootAssembly.Instance(
    dependent=ON, 
    name='Intermediate-1', 
    part=Model.parts['Intermediate-1'])

Model.rootAssembly.Instance(
    dependent=ON, 
    name='Intermediate-2', 
    part=Model.parts['Intermediate-2'])

Model.rootAssembly.translate(
    instanceList=('Intermediate-2', ), 
    vector=(0.0, yshift, 0.0))

Model.rootAssembly.InstanceFromBooleanMerge(
    domain=GEOMETRY, 
    instances=(Model.rootAssembly.instances['Intermediate-2'], Model.rootAssembly.instances['Intermediate-1']), 
    name='Intermediate-3', 
    originalInstances=DELETE)

try:
    Model.rootAssembly.features.changeKey(fromName='Intermediate-3-1', toName='Intermediate')
except:
    pass

# Cut the instance
Model.rootAssembly.InstanceFromBooleanCut(
    cuttingInstances=(Model.rootAssembly.instances['Aux'], ), 
    instanceToBeCut=Model.rootAssembly.instances['Intermediate'], 
    name='Base-Fibers', 
    originalInstances=DELETE)

Model.rootAssembly.instances.changeKey(fromName='Base-Fibers-1',toName='Base-Fibers')


#--- Mesh Definitions

# Coordinates of the cornet points of the master RVE
c1 = (p1[0]+0.5*L,p1[1]+0.5*L,0.5*L)
c2 = (p3[0]+0.5*L,p3[1]+0.5*L,0.5*L)
c3 = (p7[0]+0.5*L,p7[1]+0.5*L+yshift,0.5*L)
c4 = (p5[0]+0.5*L,p5[1]+0.5*L+yshift,0.5*L)
c5 = (p1[0]+0.5*L,p1[1]+0.5*L,-0.5*L)
c6 = (p3[0]+0.5*L,p3[1]+0.5*L,-0.5*L)
c7 = (p7[0]+0.5*L,p7[1]+0.5*L+yshift,-0.5*L)
c8 = (p5[0]+0.5*L,p5[1]+0.5*L+yshift,-0.5*L)

# Define surfaces for pressure loading (X-Y-Z) surfaces
Model.parts['Base-Fibers'].Surface(
    name='Surf-X', 
    side1Faces=Model.parts['Base-Fibers'].faces.findAt(
        ((c1[0]+0.1*(c4[0]-c1[0]),c1[1]+0.1*(c4[1]-c1[1]),0.25*L), ), 
        ((c1[0]+0.9*(c4[0]-c1[0]),c1[1]+0.9*(c4[1]-c1[1]),0.25*L), ), 
        ((c2[0]+0.1*(c3[0]-c2[0]),c2[1]+0.1*(c3[1]-c2[1]),0.25*L), ), 
        ((c2[0]+0.9*(c3[0]-c2[0]),c2[1]+0.9*(c3[1]-c2[1]),0.25*L), ), ))

Model.parts['Base-Fibers'].Surface(
    name='Surf-Y', 
    side1Faces=Model.parts['Base-Fibers'].faces.findAt(
        ((c1[0]+0.1*(c2[0]-c1[0]),c1[1]+0.1*(c2[1]-c1[1]),0.25*L), ), 
        ((c1[0]+0.9*(c2[0]-c1[0]),c1[1]+0.9*(c2[1]-c1[1]),0.25*L), ), 
        ((c4[0]+0.1*(c3[0]-c4[0]),c4[1]+0.1*(c3[1]-c4[1]),0.25*L), ), 
        ((c4[0]+0.9*(c3[0]-c4[0]),c4[1]+0.9*(c3[1]-c4[1]),0.25*L), ), ))

Model.parts['Base-Fibers'].Surface(
    name='Surf-Z', 
    side1Faces=Model.parts['Base-Fibers'].faces.findAt(
        ((c1[0]+0.1*(c3[0]-c1[0]),c1[1]+0.1*(c3[1]-c1[1]),0.5*L), ), 
        ((c1[0]+0.9*(c3[0]-c1[0]),c1[1]+0.9*(c3[1]-c1[1]),0.5*L), ), 
        ((c4[0]+0.1*(c2[0]-c4[0]),c4[1]+0.1*(c2[1]-c4[1]),0.5*L), ), 
        ((c4[0]+0.9*(c2[0]-c4[0]),c4[1]+0.9*(c2[1]-c4[1]),0.5*L), ), ))

#--------------------

# Partition faces for easier meshing
Model.parts['Base-Fibers'].PartitionFaceByShortestPath(
    faces=Model.parts['Base-Fibers'].faces.findAt(
        ((c4[0] + 0.1*(c2[0]-c4[0]), c4[1]+0.1*(c2[1]-c4[1]), 0.5), ),
        ((c4[0] + 0.9*(c2[0]-c4[0]), c4[1]+0.9*(c2[1]-c4[1]), 0.5), ),), 
    point1=Model.parts['Base-Fibers'].vertices.findAt(c2, ), 
    point2=Model.parts['Base-Fibers'].vertices.findAt(c4, ))

Model.parts['Base-Fibers'].PartitionCellByExtrudeEdge(
    cells=Model.parts['Base-Fibers'].cells, 
    edges=(Model.parts['Base-Fibers'].edges.findAt((c4[0] + 0.1*(c2[0]-c4[0]), c4[1]+0.1*(c2[1]-c4[1]), 0.5), ), ), 
    line=Model.parts['Base-Fibers'].edges.findAt((c2[0], c2[1], 0.125*L), ), 
    sense=REVERSE)

Model.parts['Base-Fibers'].PartitionCellByExtrudeEdge(
    cells=Model.parts['Base-Fibers'].cells, 
    edges=(Model.parts['Base-Fibers'].edges.findAt((c4[0] + 0.9*(c2[0]-c4[0]), c4[1]+0.9*(c2[1]-c4[1]), 0.5), ), ), 
    line=Model.parts['Base-Fibers'].edges.findAt((c2[0], c2[1], 0.125*L), ), 
    sense=REVERSE)

#--- Mesh Definitions

# Assign edge seeds
Model.parts['Base-Fibers'].seedEdgeByNumber(
    constraint=FIXED, 
    edges=Model.parts['Base-Fibers'].edges.findAt(
        ((c1[0]+(1./5.)*(c4[0]-c1[0]),c1[1]+(1./5.)*(c4[1]-c1[1]),0.5),),
        ((c1[0]+(4./5.)*(c4[0]-c1[0]),c1[1]+(4./5.)*(c4[1]-c1[1]),0.5),),
        ((c2[0]+(1./5.)*(c3[0]-c2[0]),c2[1]+(1./5.)*(c3[1]-c2[1]),0.5),),
        ((c2[0]+(4./5.)*(c3[0]-c2[0]),c2[1]+(4./5.)*(c3[1]-c2[1]),0.5),),
        ((c1[0]+(1./5.)*(c2[0]-c1[0]),c1[1]+(1./5.)*(c2[1]-c1[1]),0.5),),
        ((c1[0]+(4./5.)*(c2[0]-c1[0]),c1[1]+(4./5.)*(c2[1]-c1[1]),0.5),),
        ((c4[0]+(1./5.)*(c3[0]-c4[0]),c4[1]+(1./5.)*(c3[1]-c4[1]),0.5),),
        ((c4[0]+(4./5.)*(c3[0]-c4[0]),c4[1]+(4./5.)*(c3[1]-c4[1]),0.5),),
        ((c1[0]+(1./5.)*(c4[0]-c1[0]),c1[1]+(1./5.)*(c4[1]-c1[1]),0.0),),
        ((c1[0]+(4./5.)*(c4[0]-c1[0]),c1[1]+(4./5.)*(c4[1]-c1[1]),0.0),),
        ((c2[0]+(1./5.)*(c3[0]-c2[0]),c2[1]+(1./5.)*(c3[1]-c2[1]),0.0),),
        ((c2[0]+(4./5.)*(c3[0]-c2[0]),c2[1]+(4./5.)*(c3[1]-c2[1]),0.0),),
        ((c1[0]+(1./5.)*(c2[0]-c1[0]),c1[1]+(1./5.)*(c2[1]-c1[1]),0.0),),
        ((c1[0]+(4./5.)*(c2[0]-c1[0]),c1[1]+(4./5.)*(c2[1]-c1[1]),0.0),),
        ((c4[0]+(1./5.)*(c3[0]-c4[0]),c4[1]+(1./5.)*(c3[1]-c4[1]),0.0),),
        ((c4[0]+(4./5.)*(c3[0]-c4[0]),c4[1]+(4./5.)*(c3[1]-c4[1]),0.0),),), 
    number=numel_dir)

Model.parts['Base-Fibers'].seedEdgeByNumber(
    constraint=FIXED, 
    edges=Model.parts['Base-Fibers'].edges.findAt(
        ((c1[0],c1[1],c1[2]-0.25*L),),
        ((c2[0],c2[1],c2[2]-0.25*L),),
        ((c3[0],c3[1],c3[2]-0.25*L),),
        ((c4[0],c4[1],c4[2]-0.25*L),),), 
    number=int(0.5*numel_dir))

# Choose element type (linear or quadratic)
if use_quadratic_elements:
    Model.parts['Base-Fibers'].setElementType(
        elemTypes=(ElemType(elemCode=C3D20H, elemLibrary=STANDARD),),
        regions=tuple(Model.parts['Base-Fibers'].cells))

else:
    Model.parts['Base-Fibers'].setElementType(
        elemTypes=(ElemType(elemCode=C3D8H, elemLibrary=STANDARD),),
        regions=tuple(Model.parts['Base-Fibers'].cells))

# Generate the mesh
Model.parts['Base-Fibers'].generateMesh()

# Mirror copy and merge the instance about the XY Plane to enforce mesh symmetry at the boundary

face_x_slave_2 = Model.parts['Base-Fibers'].Set(
    name='Face-X-slave-aux-1',
    faces=Model.parts['Base-Fibers'].faces.findAt(
        ((c1[0]+(1./5.)*(c4[0]-c1[0]),c1[1]+(1./5.)*(c4[1]-c1[1]),0.25),)
        )+\
    Model.parts['Base-Fibers'].faces.findAt(
        ((c1[0]+(4./5.)*(c4[0]-c1[0]),c1[1]+(4./5.)*(c4[1]-c1[1]),0.25),)
        )
    )

face_x_master_2 = Model.parts['Base-Fibers'].Set(
    name='Face-X-master-aux-1',
    faces=Model.parts['Base-Fibers'].faces.findAt(
        ((c2[0]+(1./5.)*(c3[0]-c2[0]),c2[1]+(1./5.)*(c3[1]-c2[1]),0.25),)
        )+\
    Model.parts['Base-Fibers'].faces.findAt(
        ((c2[0]+(4./5.)*(c3[0]-c2[0]),c2[1]+(4./5.)*(c3[1]-c2[1]),0.25),)
        )
    )

del Model.rootAssembly.instances['Base-Fibers']

# Convert the part to an orphan mesh
Model.parts['Base-Fibers'].PartFromMesh(
    copySets=True, 
    name='Base-Fibers-Mesh')

prt = Model.Part(
    compressFeatureList=OFF, 
    mirrorPlane=XYPLANE, 
    name='Base-Fibers-Mesh-2',
    objectToCopy=Model.parts['Base-Fibers-Mesh'])

part_instance = Model.rootAssembly.Instance(
    dependent=ON,
    name='Base-Fibers-1',
    part=Model.parts['Base-Fibers-Mesh'])

part_instance2 = Model.rootAssembly.Instance(
    dependent=ON,
    name='Base-Fibers-2',
    part=Model.parts['Base-Fibers-Mesh-2'])

Model.rootAssembly.regenerate()

Model.rootAssembly.InstanceFromBooleanMerge(
    domain=MESH, 
    instances=tuple([inst[1] for inst in Model.rootAssembly.instances.items()]),
    mergeNodes=BOUNDARY_ONLY, 
    name='Fibers-Master', 
    nodeMergingTolerance=1e-04, 
    originalInstances=DELETE)

Model.rootAssembly.regenerate()

Model.rootAssembly.instances.changeKey(fromName='Fibers-Master-1',toName='Fibers-Master')

#--- Material and Section Definitions

Model.Material(name='Fiber')

# Linear Elastic Material Properties
Model.materials['Fiber'].Elastic(
    table=((E,nu),))

# Make a Solid Section
Model.HomogeneousSolidSection(
    material='Fiber', 
    name='Fiber-Section',
    thickness=None)

# Define a set that contains the whole part
Model.parts['Fibers-Master'].Set(
    elements=Model.parts['Fibers-Master'].elements, 
    name='Fibers-Master-Set')

# Assign the section to the part
Model.parts['Fibers-Master'].SectionAssignment(
    offset=0.0, 
    offsetField='', 
    offsetType=MIDDLE_SURFACE, 
    region=Model.parts['Fibers-Master'].sets['Fibers-Master-Set'], 
    sectionName='Fiber-Section',
    thicknessAssignment=FROM_SECTION)


Model.rootAssembly.regenerate()


#--- Set Definitions

# Master and slave nodes for the X-Face
face_x_slave_1 = Model.rootAssembly.Set(
    name='Face-X-slave-aux',
    nodes=Model.rootAssembly.instances['Fibers-Master'].sets['Face-X-slave-aux-1'].nodes)

face_x_master_1 = Model.rootAssembly.Set(
    name='Face-X-master-aux',
    nodes=Model.rootAssembly.instances['Fibers-Master'].sets['Face-X-master-aux-1'].nodes)

# Create sets for the periodic boundary conditions

# Master corner node
corner_master = Model.rootAssembly.Set(
    name = 'Corner-master',
    nodes = Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c3,
        radius=tol))

# Slave corner nodes
corner_slaves = Model.rootAssembly.Set(
    name='Corner-slaves',
    nodes=Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c1,
        radius=tol)+\
    Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c2,
        radius=tol)+\
    Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c4,
        radius=tol)+\
    Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c5,
        radius=tol)+\
    Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c6,
        radius=tol)+\
    Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c7,
        radius=tol)+\
    Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingSphere(
        center=c8,
        radius=tol)
    )

# Master and slave edge-X nodes
edge_x_master = Model.rootAssembly.Set(
    name='Edge-X-master',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c4[0]+tol,c4[1],c4[2]),
        center2=(c3[0]-tol,c3[1],c3[2]),
        radius=tol)))

edge_x_slave_1 = Model.rootAssembly.Set(
    name='Edge-X-slave_1',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c1[0]+tol,c1[1],c1[2]),
        center2=(c2[0]-tol,c2[1],c2[2]),
        radius=tol)))

edge_x_slave_2 = Model.rootAssembly.Set(
    name='Edge-X-slave_2',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c5[0]+tol,c5[1],c5[2]),
        center2=(c6[0]-tol,c6[1],c6[2]),
        radius=tol)))

edge_x_slave_3 = Model.rootAssembly.Set(
    name='Edge-X-slave_3',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c8[0]+tol,c8[1],c8[2]),
        center2=(c7[0]-tol,c7[1],c7[2]),
        radius=tol)))

edge_x_slaves = Model.rootAssembly.SetByBoolean(
    name='Edge-X-slaves',
    sets=(edge_x_slave_1,edge_x_slave_2,edge_x_slave_3),
    operation=SymbolicConstant('UNION'))

# Master and slave edge-Y nodes
edge_y_master = Model.rootAssembly.Set(
    name='Edge-Y-master',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c2[0],c2[1]+tol,c2[2]),
        center2=(c3[0],c3[1]-tol,c3[2]),
        radius=tol)))

edge_y_slave_1 = Model.rootAssembly.Set(
    name='Edge-Y-slave_1',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c1[0],c1[1]+tol,c1[2]),
        center2=(c4[0],c4[1]-tol,c4[2]),
        radius=tol)))

edge_y_slave_2 = Model.rootAssembly.Set(
    name='Edge-Y-slave_2',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c6[0],c6[1]+tol,c6[2]),
        center2=(c7[0],c7[1]-tol,c7[2]),
        radius=tol)))

edge_y_slave_3 = Model.rootAssembly.Set(
    name='Edge-Y-slave_3',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c5[0],c5[1]+tol,c5[2]),
        center2=(c8[0],c8[1]-tol,c8[2]),
        radius=tol)))

edge_y_slaves = Model.rootAssembly.SetByBoolean(
    name='Edge-Y-slaves',
    sets=(edge_y_slave_1,edge_y_slave_2,edge_y_slave_3),
    operation=SymbolicConstant('UNION'))

# Master and slave edge-Z nodes
edge_z_master = Model.rootAssembly.Set(
    name='Edge-Z-master',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c3[0],c3[1],c3[2]-tol),
        center2=(c7[0],c7[1],c7[2]+tol),
        radius=tol)))

edge_z_slave_1 = Model.rootAssembly.Set(
    name='Edge-Z-slave_1',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c1[0],c1[1],c1[2]-tol),
        center2=(c5[0],c5[1],c5[2]+tol),
        radius=tol)))

edge_z_slave_2 = Model.rootAssembly.Set(
    name='Edge-Z-slave_2',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c2[0],c2[1],c2[2]-tol),
        center2=(c6[0],c6[1],c6[2]+tol),
        radius=tol)))

edge_z_slave_3 = Model.rootAssembly.Set(
    name='Edge-Z-slave_3',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingCylinder(
        center1=(c4[0],c4[1],c4[2]-tol),
        center2=(c8[0],c8[1],c8[2]+tol),
        radius=tol)))

edge_z_slaves = Model.rootAssembly.SetByBoolean(
    name='Edge-Z-slaves',
    sets=(edge_z_slave_1,edge_z_slave_2,edge_z_slave_3),
    operation=SymbolicConstant('UNION'))

# Master and slave Face-Y nodes
face_y_slave_1 = Model.rootAssembly.Set(
    name='Face-Y-slave-1',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingBox(
        xMin=c5[0]+tol,
        yMin=c5[1]-tol,
        zMin=c5[2]+tol,
        xMax=c2[0]-tol,
        yMax=c2[1]+tol,
        zMax=c2[2]-tol,
        ),))

face_y_slave = Model.rootAssembly.SetByBoolean(
    name='Face-Y-slave',
    sets=(face_y_slave_1,edge_x_master,edge_x_slaves,edge_y_master,edge_y_slaves,edge_z_master,edge_z_slaves,corner_master,corner_slaves),
    operation=SymbolicConstant('DIFFERENCE'))

face_y_master_1 = Model.rootAssembly.Set(
    name='Face-Y-master-1',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingBox(
        xMin=c8[0]+tol,
        yMin=c8[1]-tol,
        zMin=c8[2]+tol,
        xMax=c3[0]-tol,
        yMax=c3[1]+tol,
        zMax=c3[2]-tol,
        ),))

face_y_master = Model.rootAssembly.SetByBoolean(
    name='Face-Y-master',
    sets=(face_y_master_1,edge_x_master,edge_x_slaves,edge_y_master,edge_y_slaves,edge_z_master,edge_z_slaves,corner_master,corner_slaves),
    operation=SymbolicConstant('DIFFERENCE'))

# Master and slave Face-Z nodes
face_z_slave_1 = Model.rootAssembly.Set(
    name='Face-Z-slave-aux',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingBox(
        xMin=c5[0]+tol,
        yMin=c5[1]+tol,
        zMin=c5[2]-tol,
        xMax=c7[0]-tol,
        yMax=c7[1]-tol,
        zMax=c7[2]+tol,
        ),))

face_z_slave = Model.rootAssembly.SetByBoolean(
    name='Face-Z-slave',
    sets=(face_z_slave_1,edge_x_master,edge_x_slaves,edge_y_master,edge_y_slaves,edge_z_master,edge_z_slaves,corner_master,corner_slaves),
    operation=SymbolicConstant('DIFFERENCE'))

face_z_master_1 = Model.rootAssembly.Set(
    name='Face-Z-master-aux',
    nodes=(Model.rootAssembly.instances['Fibers-Master'].nodes.getByBoundingBox(
        xMin=c1[0]+tol,
        yMin=c1[1]+tol,
        zMin=c1[2]-tol,
        xMax=c3[0]-tol,
        yMax=c3[1]-tol,
        zMax=c3[2]+tol,
        ),))

face_z_master = Model.rootAssembly.SetByBoolean(
    name='Face-Z-master',
    sets=(face_z_master_1,edge_x_master,edge_x_slaves,edge_y_master,edge_y_slaves,edge_z_master,edge_z_slaves,corner_master,corner_slaves),
    operation=SymbolicConstant('DIFFERENCE'))

face_x_slave = Model.rootAssembly.SetByBoolean(
    name='Face-X-slave',
    sets=(face_x_slave_1,edge_x_master,edge_x_slaves,edge_y_master,edge_y_slaves,edge_z_master,edge_z_slaves,corner_master,corner_slaves),
    operation=SymbolicConstant('DIFFERENCE'))

face_x_master = Model.rootAssembly.SetByBoolean(
    name='Face-X-master',
    sets=(face_x_master_1,edge_x_master,edge_x_slaves,edge_y_master,edge_y_slaves,edge_z_master,edge_z_slaves,corner_master,corner_slaves),
    operation=SymbolicConstant('DIFFERENCE'))

master_boundary = Model.rootAssembly.SetByBoolean(
    name='Master-Boundary',
    sets=(corner_slaves,corner_master,edge_x_slaves,edge_x_master,edge_y_slaves,edge_y_master,edge_z_slaves,edge_z_master,face_x_slave,face_x_master,face_y_slave,face_y_master,face_z_slave,face_z_master),
    operation=SymbolicConstant('UNION'))

# Tolerance for node based selection
ntol = 1e-6

# Lattice Vectors
LV1 = np.array([c2[0]-c1[0],c2[1]-c1[1],c2[2]-c1[2]])
LV2 = np.array([c4[0]-c1[0],c4[1]-c1[1],c4[2]-c1[2]])
LV3 = np.array([c1[0]-c5[0],c1[1]-c5[1],c1[2]-c5[2]])

# Combine all lattice vectors
LV = np.array([LV1,LV2,LV3])

'''
Mirror position identifiers:

R: Right = Positive X
L: Left  = Negative X
U: Up    = Positive Y
D: Down  = Negative Y
F: Front = Positive Z
B: Back  = Negative Z
C: Center= No effect on axis

XXX [Position in X, Position in Y, Position in Z]

if a mirror is at the center of a face then use C at the proper location

RUF = (Positive X, Positive Y, Positive Z)
RCF = (Positive X, No effect on Y, Positive Z)
CUB = (No effect on X, Positive Y, Negative Z)

A total of 6 central mirrors (Centers of each face)
A total of 12 edge mirrors (Centers of each edge)
A total of 8 corner mirrors (Corners of the master cube)

'''


# Centers of each Face
RCC = LV1
LCC = -LV1
CUC = LV2
CDC = -LV2
CCF = LV3
CCB = -LV3

# Centers of each Edge
RUC = LV1+LV2
RDC = LV1-LV2
LUC = -LV1 + LV2
LDC = -LV1-LV2

CUF = LV2 + LV3
CUB = LV2 - LV3
CDF = -LV2 + LV3
CDB = -LV2 - LV3

RCF = LV1 + LV3
RCB = LV1 - LV3
LCF = -LV1 + LV3
LCB = -LV1 - LV3

# Corners
RUF = LV1 + LV2 + LV3
RUB = LV1 + LV2 - LV3
RDF = LV1 - LV2 + LV3
RDB = LV1 - LV2 - LV3

LUF = -LV1 + LV2 + LV3
LUB = -LV1 + LV2 - LV3
LDF = -LV1 - LV2 + LV3
LDB = -LV1 - LV2 - LV3

mirror_names = ['RCC','LCC','CUC','CDC','CCF','CCB',\
                'RUC','RDC','LUC','LDC','CUF','CUB','CDF','CDB','RCF','RCB','LCF','LCB',\
                'RUF','RUB','RDF','RDB','LUF','LUB','LDF','LDB']

mirror_vectors = [RCC,LCC,CUC,CDC,CCF,CCB,\
                  RUC,RDC,LUC,LDC,CUF,CUB,CDF,CDB,RCF,RCB,LCF,LCB,\
                  RUF,RUB,RDF,RDB,LUF,LUB,LDF,LDB]

# Add the 26 additional RVEs in include_slaves is True
if include_slaves:

    # Ininialize instance list
    inst_Fibers = [Model.rootAssembly.instances['Fibers-Master']]

    # For each of the 26 cases, make the instance and translate it in space to the appropriate location
    for i, (nm,shiftv) in enumerate(zip(mirror_names,mirror_vectors)):

        H = Model.rootAssembly.Instance(
            dependent=ON, 
            name='Fibers-{}'.format(nm), 
            part=Model.parts['Fibers-Master'])

        Model.rootAssembly.translate(
            instanceList=('Fibers-{}'.format(nm), ), 
            vector=shiftv)

        inst_Fibers.append(H)  

    # Merge boundary nodes
    Model.rootAssembly.InstanceFromBooleanMerge(
        domain=MESH,
        instances=tuple(inst_Fibers),
        mergeNodes=BOUNDARY_ONLY, 
        name='Fibers',
        nodeMergingTolerance=1e-04, 
        originalInstances=DELETE)

    # Rename instance
    Model.rootAssembly.features.changeKey(
        fromName='Fibers-1',
        toName='Fibers'
)
else:
    # No slaves. Still rename instance
    Model.rootAssembly.features.changeKey(
    fromName='Fibers-Master',
    toName='Fibers')

# Regenerate the assembly
Model.rootAssembly.regenerate()

# Reference point names
RP1 = 'Reference-Point-1'
RP2 = 'Reference-Point-2'
RP3 = 'Reference-Point-3'

Model.Part(dimensionality=THREE_D, name=RP1, type=DEFORMABLE_BODY)
Model.Part(dimensionality=THREE_D, name=RP2, type=DEFORMABLE_BODY)
Model.Part(dimensionality=THREE_D, name=RP3, type=DEFORMABLE_BODY)

Model.parts[RP1].ReferencePoint(point=(0.0, 0.0, 0.0))
Model.parts[RP2].ReferencePoint(point=(0.0, 0.0, 0.0))
Model.parts[RP3].ReferencePoint(point=(0.0, 0.0, 0.0))

# Instanciate the reference points
Model.rootAssembly.Instance(
    dependent=ON, 
    name=RP1, 
    part=Model.parts[RP1]
    )

Model.rootAssembly.Instance(
    dependent=ON, 
    name=RP2, 
    part=Model.parts[RP2]
    )

Model.rootAssembly.Instance(
    dependent=ON, 
    name=RP3, 
    part=Model.parts[RP3]
    )

#Create set of reference points
Model.rootAssembly.Set(
    name=RP1,
    referencePoints=(Model.rootAssembly.instances[RP1].referencePoints[1],)
    )

Model.rootAssembly.Set(
    name=RP2, 
    referencePoints=(Model.rootAssembly.instances[RP2].referencePoints[1],)
    )

Model.rootAssembly.Set(
    name=RP3, 
    referencePoints=(Model.rootAssembly.instances[RP3].referencePoints[1],)
    )

# Initialize pair counter
pair = 0

# If include_slaves is True, first enforce the master-slave RVE conditions
# This will constrain all nodes of each slave RVE to the corresponding master node
if include_slaves:

    # Make a set with all the nodes in the master RVE
    master_nodes = Model.rootAssembly.Set(
        name = 'Master',
        nodes = (Model.rootAssembly.instances['Fibers'].nodes.getByBoundingBox(
            xMin = -0.5*L - ntol,
            yMin = -0.5*L - gap - ntol,
            zMin = -0.5*L - ntol,
            xMax = 0.5*L + ntol,
            yMax = 0.5*L + gap + ntol,
            zMax = 0.5*L + ntol),
            )
        )

    slave_sets = [master_nodes]
    sets_temp = slave_sets[:]

    for i,(nm,shiftv) in enumerate(zip(mirror_names,mirror_vectors)):

        slave_nodes = Model.rootAssembly.Set(
            name = 'node-temp',
            nodes = (Model.rootAssembly.instances['Fibers'].nodes.getByBoundingBox(
                xMin = -0.5*L - ntol + shiftv[0],
                yMin = -0.5*L - gap - ntol + shiftv[1],
                zMin = -0.5*L - ntol + shiftv[2],
                xMax = 0.5*L + ntol + shiftv[0],
                yMax = 0.5*L + gap + ntol + shiftv[1],
                zMax = 0.5*L + ntol + shiftv[2]),
            )
        ) 

        sets_temp.insert(0,slave_nodes)

        # Make sure that we don't select the same node twice
        set_actual = Model.rootAssembly.SetByBoolean(
            name = nm,
            sets = tuple(sets_temp),
            operation = SymbolicConstant('DIFFERENCE'))

        slave_sets.insert(0,set_actual)
        sets_temp = slave_sets[:]

        # For each node impose the condition
        for node in Model.rootAssembly.sets[nm].nodes:

            pair +=1

            # Coordinates of node from the left mirror
            crds = np.asarray(node.coordinates)
            crds2 = crds-shiftv

            dx = crds[0]-crds2[0]
            dy = crds[1]-crds2[1]
            dz = crds[2]-crds2[2]


            # Estimated coordinates for the master node
            set_master = Model.rootAssembly.Set(
                name='MC-Node-1-{0}'.format(pair),
                nodes=Model.rootAssembly.sets[nm].nodes.getByBoundingSphere(
                    center=crds,
                    radius=ntol))

            # Estimated coordinates for the slave nodes
            set_mirror = Model.rootAssembly.Set(
                name='MC-Node-2-{0}'.format(pair),
                nodes=Model.rootAssembly.sets['Master'].nodes.getByBoundingSphere(
                    center=crds2,
                    radius=ntol))

            # First equation
            Model.Equation(
                name='MC1-{0}'.format(pair),
                terms=(
                    ( 1.0,'MC-Node-1-{0}'.format(pair), 1),
                    (-1.0, 'MC-Node-2-{0}'.format(pair), 1),
                    (-dx, 'Reference-Point-1', 1), 
                    (-dy, 'Reference-Point-1', 2), 
                    (-dz, 'Reference-Point-1', 3)
                    )
                )

            # Second equation
            Model.Equation(
                name='MC2-{0}'.format(pair),
                terms=(
                    ( 1.0,'MC-Node-1-{0}'.format(pair), 2),
                    (-1.0, 'MC-Node-2-{0}'.format(pair), 2),
                    ( -dx,'Reference-Point-2', 1),
                    ( -dy,'Reference-Point-2', 2),
                    ( -dz,'Reference-Point-2', 3)
                    )
                )

            # Third Equation
            Model.Equation(
                name='MC3-{0}'.format(pair),
                terms=(
                    ( 1.0,'MC-Node-1-{0}'.format(pair), 3),
                    (-1.0, 'MC-Node-2-{0}'.format(pair), 3),
                    ( -dx,'Reference-Point-3', 1),
                    ( -dy,'Reference-Point-3', 2),
                    ( -dz,'Reference-Point-3', 3)
                    )
                )  

# Enforce PBC at the corner nodes
RP1,RP2,RP3,cp = PBC3D(Model,'Corner-slaves','Corner-master',LV,-2,1)

# Enforce PBC at the edge nodes
RP1,RP2,RP3,cp = PBC3D(Model,'Edge-X-slaves','Edge-X-master',LV,-1,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Edge-Y-slaves','Edge-Y-master',LV,-1,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Edge-Z-slaves','Edge-Z-master',LV,-1,cp)

# Enforce PBS at the face nodes
RP1,RP2,RP3,cp = PBC3D(Model,'Face-X-slave','Face-X-master',LV,1,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Face-Y-slave','Face-Y-master',LV,2,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Face-Z-slave','Face-Z-master',LV,3,cp)

# Eliminate Rigid body motions
Model.DisplacementBC(amplitude=UNSET, 
    createStepName='Initial', 
    distributionType=UNIFORM, 
    fieldName='', 
    localCsys=None, 
    name='BC-FIXNODE-Rigid1', 
    region=Region(nodes=Model.rootAssembly.instances['Fibers'].nodes.getByBoundingSphere(center=(c1[0]+Seedsize,c1[1]+Seedsize,0), radius=0.2*Seedsize)), 
    u1=SET, 
    u2=SET, 
    u3=SET, 
    ur1=UNSET, 
    ur2=UNSET, 
    ur3=UNSET)

Model.DisplacementBC(amplitude=UNSET, 
    createStepName='Initial', 
    distributionType=UNIFORM, 
    fieldName='', 
    localCsys=None, 
    name='BC-FIXNODE-Rigid2', 
    region=Region(nodes=Model.rootAssembly.instances['Fibers'].nodes.getByBoundingSphere(center=(c1[0]+2.5*Seedsize,c1[1]+Seedsize,0.), radius=0.3*Seedsize)), 
    u1=UNSET, 
    u2=SET, 
    u3=UNSET, 
    ur1=UNSET, 
    ur2=UNSET, 
    ur3=UNSET)

#Defining contact pairs and properties
Model.ContactProperty('Contact Interaction')

# Tangential Behavior
Model.interactionProperties['Contact Interaction'].TangentialBehavior(
    dependencies=0, 
    directionality=ISOTROPIC, 
    elasticSlipStiffness=None, 
    formulation=PENALTY, 
    fraction=0.0005, 
    maximumElasticSlip=FRACTION, 
    pressureDependency=OFF, 
    shearStressLimit=None, 
    slipRateDependency=OFF, 
    table=((0.6, ), ), 
    temperatureDependency=OFF)

# Normal Behavior
Model.interactionProperties['Contact Interaction'].NormalBehavior(
    allowSeparation=ON, 
    constraintEnforcementMethod=DEFAULT, 
    pressureOverclosure=HARD)

# Define contacts between the surfaces
Model.ContactStd(createStepName='Initial', name='Contact')

Model.interactions['Contact'].includedPairs.setValuesInStep(
    stepName='Initial',
    useAllstar=ON)

Model.interactions['Contact'].contactPropertyAssignments.appendInStep(
    assignments=((GLOBAL, SELF, 'Contact Interaction'), ), 
    stepName='Initial')


# Make step to apply pressure
Model.StaticStep(
    initialInc=0.05, 
    name='Make Contact', 
    maxInc=0.05,
    minInc=0.00000001,
    maxNumInc=1000,
    nlgeom=ON, 
    previous='Initial',
    adaptiveDampingRatio=0.05,
    stabilizationMagnitude=0.002, 
    stabilizationMethod=DISSIPATED_ENERGY_FRACTION)

# Make step to apply pressure
Model.StaticStep(
    initialInc=5e-3, 
    name='Apply Pressure', 
    maxInc=1e-2,
    minInc=1e-10,
    maxNumInc=10000,
    nlgeom=ON, 
    previous='Make Contact',
    adaptiveDampingRatio=0.05,
    stabilizationMagnitude=0.002, 
    stabilizationMethod=DISSIPATED_ENERGY_FRACTION)

# Request history and field output data
Model.FieldOutputRequest(createStepName='Make Contact', name='F-Output-0', variables=PRESELECT)

Model.HistoryOutputRequest(createStepName='Make Contact', name='H-Output-0', variables=PRESELECT)

Model.HistoryOutputRequest(
    createStepName='Apply Pressure', 
    name='History-RP1', 
    rebar=EXCLUDE, 
    region=Model.rootAssembly.sets['Reference-Point-1'], 
    sectionPoints=DEFAULT, 
    variables=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'TF1', 'TF2', 'TF3'))

Model.HistoryOutputRequest(
    createStepName='Apply Pressure', 
    name='History-RP2', 
    rebar=EXCLUDE, 
    region=Model.rootAssembly.sets['Reference-Point-2'], 
    sectionPoints=DEFAULT, 
    variables=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'TF1', 'TF2', 'TF3'))


Model.HistoryOutputRequest(
    createStepName='Apply Pressure', 
    name='History-RP3', 
    rebar=EXCLUDE, 
    region=Model.rootAssembly.sets['Reference-Point-3'], 
    sectionPoints=DEFAULT, 
    variables=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'TF1', 'TF2', 'TF3'))

# Request restart capability during the pressure step (5 points to restart from)
Model.steps['Apply Pressure'].Restart(
    frequency=0, 
    numberIntervals=5, 
    overlay=OFF, 
    timeMarks=OFF)


Model.FieldOutputRequest(createStepName='Apply Pressure', name='F-Output-1', variables=PRESELECT)
Model.fieldOutputRequests['F-Output-1'].setValues(numIntervals=300,timeMarks=OFF)

Model.HistoryOutputRequest(createStepName='Apply Pressure', name='H-Output-1', variables=PRESELECT)
Model.historyOutputRequests['H-Output-1'].setValues(numIntervals=300, timeMarks=OFF)

# Initialize contact by eliminating the gap delta
Model.DisplacementBC(
    amplitude=UNSET, 
    createStepName='Make Contact', 
    distributionType=UNIFORM, 
    fieldName='', fixed=OFF, 
    localCsys=None, 
    name='BC-Periodic-1', 
    region=Region(referencePoints=(Model.rootAssembly.instances[RP1].referencePoints[1],)), 
    u1=gap,
    u2=0.0,
    u3=0.0)

Model.DisplacementBC(
    amplitude=UNSET, 
    createStepName='Make Contact', 
    distributionType=UNIFORM, 
    fieldName='', fixed=OFF, 
    localCsys=None, 
    name='BC-Periodic-2', 
    region=Region(referencePoints=(Model.rootAssembly.instances[RP2].referencePoints[1],)),  
    u1=0,
    u2=gap,
    u3=0)

Model.DisplacementBC(
    amplitude=UNSET, 
    createStepName='Make Contact', 
    distributionType=UNIFORM, 
    fieldName='', fixed=OFF, 
    localCsys=None, 
    name='BC-Periodic-3', 
    region=Region(referencePoints=(Model.rootAssembly.instances[RP3].referencePoints[1],)),  
    u1=0,
    u2=0,
    u3=0)

# Deactivate the BC in the pressure step
Model.boundaryConditions['BC-Periodic-1'].deactivate('Apply Pressure')
Model.boundaryConditions['BC-Periodic-2'].deactivate('Apply Pressure')
Model.boundaryConditions['BC-Periodic-3'].deactivate('Apply Pressure')

# Apply pressure at the surfaces
Model.Pressure(
    amplitude=UNSET, 
    createStepName='Apply Pressure', 
    distributionType=UNIFORM, 
    field='', 
    magnitude=p, 
    name='PX', 
    region=Model.rootAssembly.instances['Fibers'].surfaces['Surf-X'])

Model.Pressure(
    amplitude=UNSET, 
    createStepName='Apply Pressure', 
    distributionType=UNIFORM, 
    field='', 
    magnitude=p, 
    name='PY', 
    region=Model.rootAssembly.instances['Fibers'].surfaces['Surf-Y'])


Model.Pressure(
    amplitude=UNSET, 
    createStepName='Apply Pressure', 
    distributionType=UNIFORM, 
    field='', 
    magnitude=p, 
    name='PZ', 
    region=Model.rootAssembly.instances['Fibers'].surfaces['Surf-Z'])


# Make job
mdb.Job(
    atTime=None, 
    contactPrint=OFF, 
    description='', 
    echoPrint=OFF,
    explicitPrecision=SINGLE,  
    getMemoryFromAnalysis=True,
    historyPrint=OFF, 
    memory=100, 
    memoryUnits=PERCENTAGE, 
    model=ModelName, 
    modelPrint=OFF,
    multiprocessingMode=THREADS, 
    name=JobName, 
    nodalOutputPrecision=SINGLE,
    numCpus=ncpus, 
    numDomains=ncpus, 
    numGPUs=0, 
    queue=None, 
    scratch='', 
    type=ANALYSIS, 
    userSubroutine='', 
    waitHours=0, 
    waitMinutes=0)

mdb.saveAs(pathName=Savename)

if SUBMIT:
    mdb.jobs[JobName].submit(consistencyChecking=OFF)