__author__ = "N. Vasios"
__email__ = "vasios@g.harvard.edu"

'''
Numerical Analysis of Periodic Laminar and Fibrous media undergoing aJamming Transition
N. Vasios, B. Aktas, Y. Narang, R. Howe, K. Bertoldi

Abaqus python script for the numerical analysis of laminar RVEs
'''

from part import *
from material import *
from section import *
from assembly import *
from step import *
from interaction import *
from load import *
from mesh import *
from optimization import *
from job import *
from sketch import *
from visualization import *
from connectorBehavior import *
import numpy as np

execfile("utilities.py")

## Naming Parts, Materials, Instances, Jobs etc
ModelName = 'Model-1'


#--------------------------------------------------------
# USER DEFINED PARAMETERS
#

# Model dimensions
L = 1.0 # Fiber width, and depth

# Gap between fibers and tolerance for node-based selection
gap = 1e-2
tol = 1e-5


# Number of finite elements per dimension
numel_dir = 5

# Material Parameters
E,nu = 2000, 0.42  # Data for polypropelene plastic

# Value of pressure to be applied in MPa
p = 0.101

# Number of cpus
ncpus = 8 
SUBMIT = True

# Boolean to include slave RVEs
include_slaves = False

# Boolean to use quadratic elements
use_quadratic_elements = False

#--------------------------------------------------------

# Job name and cae file name
JobName = 'Laminar_3D'
Savename = JobName + '.cae'

# Global seed size corresponding to the number of elements per dimension chosen
Seedsize = L/numel_dir

# Change directory of abaqus to the directory of the python script
os.chdir(os.path.dirname(os.path.realpath('__file__')))

#--------------------------------------------------------------------------------------

# Use coordinates to represent geometric selections
session.journalOptions.setValues(replayGeometry=COORDINATE, recoverGeometry=COORDINATE)

# Initialize the Model Database
Mdb()

# Get the model of interest
Model = mdb.models['Model-1']

# Use a cartesian coordinate system
Model.rootAssembly.DatumCsysByDefault(CARTESIAN)


#--- Part Definitions

# Sketch the 2D cross-section of the part
Model.ConstrainedSketch(name='__profile__', sheetSize=L)


'''

Schematic of the 2D cross-section of the part

(-)-------------(p4)
 |                 |
 |                 |
(p3)-------------(--)


(--)-------------(p2)
 |                 |
 |                 |
(p1)-------------(--)

'''

# The coordinates of the points of interest

p1 = (0,-0.5*L-gap,0)

p2 = (0.5*L,-gap,0)


# Rectangle 1
Model.sketches['__profile__'].rectangle(
    point1=p1[:-1], 
    point2=p2[:-1])

# Make the deformable part
Model.Part(
    dimensionality=THREE_D, 
    name='Base-Layers', 
    type=DEFORMABLE_BODY)

# Extrude the 2D cross-section by half the depth of the actual part (then mirror)
Model.parts['Base-Layers'].BaseSolidExtrude(
    depth=0.5*L, 
    sketch=Model.sketches['__profile__'])

# Delete the sketch
del Model.sketches['__profile__']

# Define surfaces for pressure loading (X-Y-Z) surfaces
Model.parts['Base-Layers'].Surface(
    name='Surf-X', 
    side1Faces=Model.parts['Base-Layers'].faces.findAt(
        ((p2[0],p1[1]+0.5*(p2[1]-p1[1]),0.25*L), ),  ))

Model.parts['Base-Layers'].Surface(
    name='Surf-Y', 
    side1Faces=Model.parts['Base-Layers'].faces.findAt(
        ((p1[0] + 0.5*(p2[0]-p1[0]),p1[1],0.25*L), ),  ))

Model.parts['Base-Layers'].Surface(
    name='Surf-Z', 
    side1Faces=Model.parts['Base-Layers'].faces.findAt(
        ((p1[0]+0.5*(p2[0]-p1[0]),p1[1]+0.5*(p2[1]-p1[1]),0.5*L), ),  ))

#--- Mesh Definitions

# Assign global seeds
Model.parts['Base-Layers'].seedPart(
    deviationFactor=0.1, 
    minSizeFactor=0.1, 
    size=Seedsize)

# Choose element type (linear or quadratic)
if use_quadratic_elements:
    Model.parts['Base-Layers'].setElementType(
        elemTypes=(ElemType(elemCode=C3D20H, elemLibrary=STANDARD),),
        regions=tuple(Model.parts['Base-Layers'].cells))

else:
    Model.parts['Base-Layers'].setElementType(
        elemTypes=(ElemType(elemCode=C3D8H, elemLibrary=STANDARD),),
        regions=tuple(Model.parts['Base-Layers'].cells))


# Generate the mesh
Model.parts['Base-Layers'].generateMesh()


#--- Material and Section Definitions

Model.Material(name='Layer')

# Linear Elastic Material Properties
Model.materials['Layer'].Elastic(
    table=((E,nu),))

# Make a Solid Section
Model.HomogeneousSolidSection(
    material='Layer', 
    name='Layer-Section',
    thickness=None)

# Define a set that contains the whole part
Model.parts['Base-Layers'].Set(
    cells=Model.parts['Base-Layers'].cells, 
    name='Base-Layers-Set')

# Assign the section to the part
Model.parts['Base-Layers'].SectionAssignment(
    offset=0.0, 
    offsetField='', 
    offsetType=MIDDLE_SURFACE, 
    region=Model.parts['Base-Layers'].sets['Base-Layers-Set'], 
    sectionName='Layer-Section',
    thicknessAssignment=FROM_SECTION)


#--- Instancing and Set Definitions

Model.parts['Base-Layers'].PartFromMesh(
    copySets=True, 
    name='Base-Layers-Mesh')

# Delete the original part
del Model.parts['Base-Layers']

# Rename the orphan mesh part
Model.parts.changeKey(fromName='Base-Layers-Mesh',toName='Base-Layers-0')

# Make the base instance
Model.rootAssembly.Instance(
    dependent=ON, 
    name='Base-Layers-0', 
    part=Model.parts['Base-Layers-0'])

# Mirror copy and merge the instance about the 3 principal planes to enforce mesh symmetry at the boundary
planes = [YZPLANE, XZPLANE, XYPLANE]

for i,plane in enumerate(planes):

    # Make a new part mirrored about a plane
    prt = Model.Part(
        compressFeatureList=ON, 
        mirrorPlane=plane, 
        name='Base-Layers-{0}'.format(i+1),
        objectToCopy=Model.parts['Base-Layers-{}'.format(i)])

    # Instanciate the part
    part_instance = Model.rootAssembly.Instance(
        dependent=ON,
        name='Base-Layers-{0}'.format(i+1),
        part=Model.parts['Base-Layers-{}'.format(i+1)])

    # Regenerate the assembly
    Model.rootAssembly.regenerate()

    # Merge and repeat
    Model.rootAssembly.InstanceFromBooleanMerge(
        domain=MESH, 
        instances=tuple([inst[1] for inst in Model.rootAssembly.instances.items()]),
        mergeNodes=BOUNDARY_ONLY, 
        name='Base-Layers-{0}'.format(i+1), 
        nodeMergingTolerance=1e-04, 
        originalInstances=DELETE)

    # Delete the auxiliary mirrored part
    del Model.parts['Base-Layers-{}'.format(i)]

# Rename the merged geometry
Model.rootAssembly.features.changeKey(
    fromName='Base-Layers-{0}-1'.format(i+1),
    toName='Layers-Master')

Model.parts.changeKey(fromName='Base-Layers-{0}'.format(i+1),toName='Layers-Master')

# Regenerate the assembly
Model.rootAssembly.regenerate()

# Create sets for the periodic boundary conditions

# Master corner node
corner_master = Model.rootAssembly.Set(
    name = 'Corner-master',
    nodes = Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(0.5*L,0.5*L+gap,0.5*L),
        radius=tol))

# Slave corner nodes
corner_slaves = Model.rootAssembly.Set(
    name='Corner-slaves',
    nodes=Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(-0.5*L,-0.5*L-gap,-0.5*L),
        radius=tol)+\
    Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(0.5*L,-0.5*L-gap,-0.5*L),
        radius=tol)+\
    Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(0.5*L,0.5*L+gap,-0.5*L),
        radius=tol)+\
    Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(-0.5*L,0.5*L+gap,-0.5*L),
        radius=tol)+\
    Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(-0.5*L,-0.5*L-gap,0.5*L),
        radius=tol)+\
    Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(0.5*L,-0.5*L-gap,0.5*L),
        radius=tol)+\
    Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingSphere(
        center=(-0.5*L,0.5*L+gap,0.5*L),
        radius=tol)
    )

# Master Edge nodes - X
edge_x_master = Model.rootAssembly.Set(
    name='Edge-X-master',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L+tol,0.5*L+gap,0.5*L),
        center2=(0.5*L-tol,0.5*L+gap,0.5*L),
        radius=tol)))

# Slave Edge nodes - X vol 1
edge_x_slave_1 = Model.rootAssembly.Set(
    name='Edge-X-slave_1',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L+tol,-0.5*L-gap,0.5*L),
        center2=(0.5*L-tol,-0.5*L-gap,0.5*L),
        radius=tol)))

# Slave Edge nodes - X vol 2
edge_x_slave_2 = Model.rootAssembly.Set(
    name='Edge-X-slave_2',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L+tol,0.5*L+gap,-0.5*L),
        center2=(0.5*L-tol,0.5*L+gap,-0.5*L),
        radius=tol)))

# Slave Edge nodes - X vol 3
edge_x_slave_3 = Model.rootAssembly.Set(
    name='Edge-X-slave_3',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L+tol,-0.5*L-gap,-0.5*L),
        center2=(0.5*L-tol,-0.5*L-gap,-0.5*L),
        radius=tol)))

# Merge all slave edge nodes - X
edge_x_slaves = Model.rootAssembly.SetByBoolean(
    name='Edge-X-slaves',
    sets=(edge_x_slave_1,edge_x_slave_2,edge_x_slave_3),
    operation=SymbolicConstant('UNION'))

#  Master Edge nodes - Y
edge_y_master = Model.rootAssembly.Set(
    name='Edge-Y-master',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(0.5*L,-0.5*L-gap+tol,0.5*L),
        center2=(0.5*L,0.5*L+gap-tol,0.5*L),
        radius=tol)))

# Slave Edge nodes - Y vol 1
edge_y_slave_1 = Model.rootAssembly.Set(
    name='Edge-Y-slave_1',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L,-0.5*L-gap+tol,0.5*L),
        center2=(-0.5*L,0.5*L+gap-tol,0.5*L),
        radius=tol)))

# Slave Edge nodes - Y vol 2
edge_y_slave_2 = Model.rootAssembly.Set(
    name='Edge-Y-slave_2',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(0.5*L,-0.5*L-gap+tol,-0.5*L),
        center2=(0.5*L,0.5*L+gap-tol,-0.5*L),
        radius=tol)))

# Slave Edge nodes - Y vol 3
edge_y_slave_3 = Model.rootAssembly.Set(
    name='Edge-Y-slave_3',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L,-0.5*L-gap+tol,-0.5*L),
        center2=(-0.5*L,0.5*L+gap-tol,-0.5*L),
        radius=tol)))

# Merge all slave edge nodes - Y
edge_y_slaves = Model.rootAssembly.SetByBoolean(
    name='Edge-Y-slaves',
    sets=(edge_y_slave_1,edge_y_slave_2,edge_y_slave_3),
    operation=SymbolicConstant('UNION'))

#  Master Edge nodes - Z
edge_z_master = Model.rootAssembly.Set(
    name='Edge-Z-master',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(0.5*L,0.5*L+gap,-0.5*L+tol),
        center2=(0.5*L,0.5*L+gap,0.5*L-tol),
        radius=tol)))

# Slave Edge nodes - Z vol 1
edge_z_slave_1 = Model.rootAssembly.Set(
    name='Edge-Z-slave_1',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L,0.5*L+gap,-0.5*L+tol),
        center2=(-0.5*L,0.5*L+gap,0.5*L-tol),
        radius=tol)))

# Slave Edge nodes - Z vol 2
edge_z_slave_2 = Model.rootAssembly.Set(
    name='Edge-Z-slave_2',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(0.5*L,-0.5*L-gap,-0.5*L+tol),
        center2=(0.5*L,-0.5*L-gap,0.5*L-tol),
        radius=tol)))

# Slave Edge nodes - Z vol 3
edge_z_slave_3 = Model.rootAssembly.Set(
    name='Edge-Z-slave_3',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingCylinder(
        center1=(-0.5*L,-0.5*L-gap,-0.5*L+tol),
        center2=(-0.5*L,-0.5*L-gap,0.5*L-tol),
        radius=tol)))

# Merge all slave edge nodes - Z
edge_z_slaves = Model.rootAssembly.SetByBoolean(
    name='Edge-Z-slaves',
    sets=(edge_z_slave_1,edge_z_slave_2,edge_z_slave_3),
    operation=SymbolicConstant('UNION'))

# Slave nodes on X-face
face_x_slave = Model.rootAssembly.Set(
    name='Face-X-slave',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingBox(
        xMin=-0.5*L-tol,
        yMin=-0.5*L-gap+tol,
        zMin=-0.5*L+tol,
        xMax=-0.5*L+tol,
        yMax=0.5*L+gap-tol,
        zMax=0.5*L-tol,
        ),))

# Master nodes on X-face
face_x_master = Model.rootAssembly.Set(
    name='Face-X-master',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingBox(
        xMin=0.5*L-tol,
        yMin=-0.5*L-gap+tol,
        zMin=-0.5*L+tol,
        xMax=0.5*L+tol,
        yMax=0.5*L+gap-tol,
        zMax=0.5*L-tol,
        ),))

# Slave nodes on Y-face
face_y_slave = Model.rootAssembly.Set(
    name='Face-Y-slave',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingBox(
        xMin=-0.5*L+tol,
        yMin=-0.5*L-gap-tol,
        zMin=-0.5*L+tol,
        xMax=0.5*L-tol,
        yMax=-0.5*L-gap+tol,
        zMax=0.5*L-tol,
        ),))

# Master nodes on Y-face
face_y_master = Model.rootAssembly.Set(
    name='Face-Y-master',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingBox(
        xMin=-0.5*L+tol,
        yMin=0.5*L+gap-tol,
        zMin=-0.5*L+tol,
        xMax=0.5*L-tol,
        yMax=0.5*L+gap+tol,
        zMax=0.5*L-tol,
        ),))

# Slave nodes on Z-face
face_z_slave = Model.rootAssembly.Set(
    name='Face-Z-slave',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingBox(
        xMin=-0.5*L+tol,
        yMin=-0.5*L-gap+tol,
        zMin=-0.5*L-tol,
        xMax=0.5*L-tol,
        yMax=0.5*L+gap-tol,
        zMax=-0.5*L+tol,
        ),))

# Master nodes on Z-face
face_z_master = Model.rootAssembly.Set(
    name='Face-Z-master',
    nodes=(Model.rootAssembly.instances['Layers-Master'].nodes.getByBoundingBox(
        xMin=-0.5*L+tol,
        yMin=-0.5*L-gap+tol,
        zMin=0.5*L-tol,
        xMax=0.5*L-tol,
        yMax=0.5*L+gap-tol,
        zMax=0.5*L+tol,
        ),))

# Create a set containing all the nodes involved in the periodic boundary conditions (slaves and master)
master_boundary = Model.rootAssembly.SetByBoolean(
    name='Master-Boundary',
    sets=(corner_slaves,corner_master,edge_x_slaves,edge_x_master,edge_y_slaves,edge_y_master,edge_z_slaves,edge_z_master,face_x_slave,face_x_master,face_y_slave,face_y_master,face_z_slave,face_z_master),
    operation=SymbolicConstant('UNION'))

# Tolerance for node based selection
ntol = 1e-6

# Lattice Vectors
LV1 = np.array([L,0,0])
LV2 = np.array([0,L+2*gap,0])
LV3 = np.array([0,0,L])

# Combine all lattice vectors
LV = np.array([LV1,LV2,LV3])

'''
Mirror position identifiers:

R: Right = Positive X
L: Left  = Negative X
U: Up    = Positive Y
D: Down  = Negative Y
F: Front = Positive Z
B: Back  = Negative Z
C: Center= No effect on axis

XXX [Position in X, Position in Y, Position in Z]

if a mirror is at the center of a face then use C at the proper location

RUF = (Positive X, Positive Y, Positive Z)
RCF = (Positive X, No effect on Y, Positive Z)
CUB = (No effect on X, Positive Y, Negative Z)

A total of 6 central mirrors (Centers of each face)
A total of 12 edge mirrors (Centers of each edge)
A total of 8 corner mirrors (Corners of the master cube)

'''

# Centers of each Face
RCC = LV1
LCC = -LV1
CUC = LV2
CDC = -LV2
CCF = LV3
CCB = -LV3

# Centers of each Edge
RUC = LV1+LV2
RDC = LV1-LV2
LUC = -LV1 + LV2
LDC = -LV1-LV2

CUF = LV2 + LV3
CUB = LV2 - LV3
CDF = -LV2 + LV3
CDB = -LV2 - LV3

RCF = LV1 + LV3
RCB = LV1 - LV3
LCF = -LV1 + LV3
LCB = -LV1 - LV3

# Corners
RUF = LV1 + LV2 + LV3
RUB = LV1 + LV2 - LV3
RDF = LV1 - LV2 + LV3
RDB = LV1 - LV2 - LV3

LUF = -LV1 + LV2 + LV3
LUB = -LV1 + LV2 - LV3
LDF = -LV1 - LV2 + LV3
LDB = -LV1 - LV2 - LV3

mirror_names = ['RCC','LCC','CUC','CDC','CCF','CCB',\
                'RUC','RDC','LUC','LDC','CUF','CUB','CDF','CDB','RCF','RCB','LCF','LCB',\
                'RUF','RUB','RDF','RDB','LUF','LUB','LDF','LDB']

mirror_vectors = [RCC,LCC,CUC,CDC,CCF,CCB,\
                  RUC,RDC,LUC,LDC,CUF,CUB,CDF,CDB,RCF,RCB,LCF,LCB,\
                  RUF,RUB,RDF,RDB,LUF,LUB,LDF,LDB]

# Add the 26 additional RVEs in include_slaves is True
if include_slaves:

    # Ininialize instance list
    inst_Layers = [Model.rootAssembly.instances['Layers-Master']]

    # For each of the 26 cases, make the instance and translate it in space to the appropriate location
    for i, (nm,shiftv) in enumerate(zip(mirror_names,mirror_vectors)):

        H = Model.rootAssembly.Instance(
            dependent=ON, 
            name='Layers-{}'.format(nm), 
            part=Model.parts['Layers-Master'])

        Model.rootAssembly.translate(
            instanceList=('Layers-{}'.format(nm), ), 
            vector=shiftv)

        inst_Layers.append(H)  

    # Merge boundary nodes
    Model.rootAssembly.InstanceFromBooleanMerge(
        domain=MESH,
        instances=tuple(inst_Layers),
        mergeNodes=BOUNDARY_ONLY, 
        name='Layers',
        nodeMergingTolerance=1e-04, 
        originalInstances=DELETE)

    # Rename instance
    Model.rootAssembly.features.changeKey(
        fromName='Layers-1',
        toName='Layers'
)
else:
    # No slaves. Still rename instance
    Model.rootAssembly.features.changeKey(
    fromName='Layers-Master',
    toName='Layers')

# Regenerate the assembly
Model.rootAssembly.regenerate()

# Reference point names
RP1 = 'Reference-Point-1'
RP2 = 'Reference-Point-2'
RP3 = 'Reference-Point-3'

# Make 3 parts and assign each one to the corresponding RP
Model.Part(dimensionality=THREE_D, name=RP1, type=DEFORMABLE_BODY)
Model.Part(dimensionality=THREE_D, name=RP2, type=DEFORMABLE_BODY)
Model.Part(dimensionality=THREE_D, name=RP3, type=DEFORMABLE_BODY)

Model.parts[RP1].ReferencePoint(point=(0.0, 0.0, 0.0))
Model.parts[RP2].ReferencePoint(point=(0.0, 0.0, 0.0))
Model.parts[RP3].ReferencePoint(point=(0.0, 0.0, 0.0))

# Instanciate the reference points
Model.rootAssembly.Instance(
    dependent=ON, 
    name=RP1, 
    part=Model.parts[RP1]
    )

Model.rootAssembly.Instance(
    dependent=ON, 
    name=RP2, 
    part=Model.parts[RP2]
    )

Model.rootAssembly.Instance(
    dependent=ON, 
    name=RP3, 
    part=Model.parts[RP3]
    )

#Create set of reference points
Model.rootAssembly.Set(
    name=RP1,
    referencePoints=(Model.rootAssembly.instances[RP1].referencePoints[1],)
    )

Model.rootAssembly.Set(
    name=RP2, 
    referencePoints=(Model.rootAssembly.instances[RP2].referencePoints[1],)
    )

Model.rootAssembly.Set(
    name=RP3, 
    referencePoints=(Model.rootAssembly.instances[RP3].referencePoints[1],)
    )

# Initialize pair counter
pair = 0


# If include_slaves is True, first enforce the master-slave RVE conditions
# This will constrain all nodes of each slave RVE to the corresponding master node
if include_slaves:

    # Make a set with all the nodes in the master RVE
    master_nodes = Model.rootAssembly.Set(
        name = 'Master',
        nodes = (Model.rootAssembly.instances['Layers'].nodes.getByBoundingBox(
            xMin = -0.5*L - ntol,
            yMin = -0.5*L - gap - ntol,
            zMin = -0.5*L - ntol,
            xMax = 0.5*L + ntol,
            yMax = 0.5*L + gap + ntol,
            zMax = 0.5*L + ntol),
            )
        )

    slave_sets = [master_nodes]
    sets_temp = slave_sets[:]

    for i,(nm,shiftv) in enumerate(zip(mirror_names,mirror_vectors)):

        # Get the slave nodes of the i-th slave RVE
        slave_nodes = Model.rootAssembly.Set(
            name = 'node-temp',
            nodes = (Model.rootAssembly.instances['Layers'].nodes.getByBoundingBox(
                xMin = -0.5*L - ntol + shiftv[0],
                yMin = -0.5*L - gap - ntol + shiftv[1],
                zMin = -0.5*L - ntol + shiftv[2],
                xMax = 0.5*L + ntol + shiftv[0],
                yMax = 0.5*L + gap + ntol + shiftv[1],
                zMax = 0.5*L + ntol + shiftv[2]),
            )
        ) 

        sets_temp.insert(0,slave_nodes)

        # Make sure that we don't select the same node twice
        set_actual = Model.rootAssembly.SetByBoolean(
            name = nm,
            sets = tuple(sets_temp),
            operation = SymbolicConstant('DIFFERENCE'))

        slave_sets.insert(0,set_actual)
        sets_temp = slave_sets[:]

        # For each node impose the condition
        for node in Model.rootAssembly.sets[nm].nodes:

            pair +=1

            # Coordinates of node from the left mirror
            crds = np.asarray(node.coordinates)
            crds2 = crds-shiftv

            dx = crds[0]-crds2[0]
            dy = crds[1]-crds2[1]
            dz = crds[2]-crds2[2]


            # Estimated coordinates for the master node
            set_master = Model.rootAssembly.Set(
                name='MC-Node-1-{0}'.format(pair),
                nodes=Model.rootAssembly.sets[nm].nodes.getByBoundingSphere(
                    center=crds,
                    radius=ntol))

            # Estimated coordinates for the slave nodes
            set_mirror = Model.rootAssembly.Set(
                name='MC-Node-2-{0}'.format(pair),
                nodes=Model.rootAssembly.sets['Master'].nodes.getByBoundingSphere(
                    center=crds2,
                    radius=ntol))

            # First equation
            Model.Equation(
                name='MC1-{0}'.format(pair),
                terms=(
                    ( 1.0,'MC-Node-1-{0}'.format(pair), 1),
                    (-1.0, 'MC-Node-2-{0}'.format(pair), 1),
                    (-dx, 'Reference-Point-1', 1), 
                    (-dy, 'Reference-Point-1', 2), 
                    (-dz, 'Reference-Point-1', 3)
                    )
                )

            # Second Equation
            Model.Equation(
                name='MC2-{0}'.format(pair),
                terms=(
                    ( 1.0,'MC-Node-1-{0}'.format(pair), 2),
                    (-1.0, 'MC-Node-2-{0}'.format(pair), 2),
                    ( -dx,'Reference-Point-2', 1),
                    ( -dy,'Reference-Point-2', 2),
                    ( -dz,'Reference-Point-2', 3)
                    )
                )

            # Third Equation
            Model.Equation(
                name='MC3-{0}'.format(pair),
                terms=(
                    ( 1.0,'MC-Node-1-{0}'.format(pair), 3),
                    (-1.0, 'MC-Node-2-{0}'.format(pair), 3),
                    ( -dx,'Reference-Point-3', 1),
                    ( -dy,'Reference-Point-3', 2),
                    ( -dz,'Reference-Point-3', 3)
                    )
                )  

# Enforce PBC at the corner nodes
RP1,RP2,RP3,cp = PBC3D(Model,'Corner-slaves','Corner-master',LV,-2,1)

# Enforce PBC at the edge nodes
RP1,RP2,RP3,cp = PBC3D(Model,'Edge-X-slaves','Edge-X-master',LV,-1,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Edge-Y-slaves','Edge-Y-master',LV,-1,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Edge-Z-slaves','Edge-Z-master',LV,-1,cp)

# Enforce PBS at the face nodes
RP1,RP2,RP3,cp = PBC3D(Model,'Face-X-slave','Face-X-master',LV,1,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Face-Y-slave','Face-Y-master',LV,2,cp)
RP1,RP2,RP3,cp = PBC3D(Model,'Face-Z-slave','Face-Z-master',LV,3,cp)

# Eliminate Rigid body motions
Model.DisplacementBC(amplitude=UNSET, 
    createStepName='Initial', 
    distributionType=UNIFORM, 
    fieldName='', 
    localCsys=None, 
    name='BC-FIXNODE-Rigid1', 
    region=Region(nodes=Model.rootAssembly.instances['Layers'].nodes.getByBoundingSphere(center=(0,-gap-Seedsize,0), radius=0.5*Seedsize)), 
    u1=SET, 
    u2=SET, 
    u3=SET, 
    ur1=UNSET, 
    ur2=UNSET, 
    ur3=UNSET)


#Defining contact pairs and properties
Model.ContactProperty('Contact Interaction')

# Tangential Behavior
Model.interactionProperties['Contact Interaction'].TangentialBehavior(
    dependencies=0, 
    directionality=ISOTROPIC, 
    elasticSlipStiffness=None, 
    formulation=PENALTY, 
    fraction=0.0005, 
    maximumElasticSlip=FRACTION, 
    pressureDependency=OFF, 
    shearStressLimit=None, 
    slipRateDependency=OFF, 
    table=((0.6, ), ), 
    temperatureDependency=OFF)

# Normal Behavior
Model.interactionProperties['Contact Interaction'].NormalBehavior(
    allowSeparation=ON, 
    constraintEnforcementMethod=DEFAULT, 
    pressureOverclosure=HARD)

# Define contacts between the surfaces
Model.ContactStd(createStepName='Initial', name='Contact')

Model.interactions['Contact'].includedPairs.setValuesInStep(
    stepName='Initial',
    useAllstar=ON)

Model.interactions['Contact'].contactPropertyAssignments.appendInStep(
    assignments=((GLOBAL, SELF, 'Contact Interaction'), ), 
    stepName='Initial')


# Make step to apply pressure
Model.StaticStep(
    initialInc=0.05, 
    name='Make Contact', 
    maxInc=0.05,
    minInc=0.00000001,
    maxNumInc=1000,
    nlgeom=ON, 
    previous='Initial',
    adaptiveDampingRatio=0.05,
    stabilizationMagnitude=0.0002, 
    stabilizationMethod=DISSIPATED_ENERGY_FRACTION)

# Make step to apply pressure
Model.StaticStep(
    initialInc=5e-3, 
    name='Apply Pressure', 
    maxInc=1e-2,
    minInc=1e-10,
    maxNumInc=10000,
    nlgeom=ON, 
    previous='Make Contact',
    adaptiveDampingRatio=0.05,
    stabilizationMagnitude=0.0002, 
    stabilizationMethod=DISSIPATED_ENERGY_FRACTION)

# Request history and field output data
Model.FieldOutputRequest(createStepName='Make Contact', name='F-Output-0', variables=PRESELECT)

Model.HistoryOutputRequest(createStepName='Make Contact', name='H-Output-0', variables=PRESELECT)

Model.HistoryOutputRequest(
    createStepName='Apply Pressure', 
    name='History-RP1', 
    rebar=EXCLUDE, 
    region=Model.rootAssembly.sets['Reference-Point-1'], 
    sectionPoints=DEFAULT, 
    variables=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'TF1', 'TF2', 'TF3'))

Model.HistoryOutputRequest(
    createStepName='Apply Pressure', 
    name='History-RP2', 
    rebar=EXCLUDE, 
    region=Model.rootAssembly.sets['Reference-Point-2'], 
    sectionPoints=DEFAULT, 
    variables=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'TF1', 'TF2', 'TF3'))


Model.HistoryOutputRequest(
    createStepName='Apply Pressure', 
    name='History-RP3', 
    rebar=EXCLUDE, 
    region=Model.rootAssembly.sets['Reference-Point-3'], 
    sectionPoints=DEFAULT, 
    variables=('U1', 'U2', 'U3', 'RF1', 'RF2', 'RF3', 'TF1', 'TF2', 'TF3'))


Model.steps['Apply Pressure'].Restart(
    frequency=0, 
    numberIntervals=25, 
    overlay=OFF, 
    timeMarks=OFF)


Model.FieldOutputRequest(createStepName='Apply Pressure', name='F-Output-1', variables=PRESELECT)
Model.fieldOutputRequests['F-Output-1'].setValues(numIntervals=300,timeMarks=OFF)

Model.HistoryOutputRequest(createStepName='Apply Pressure', name='H-Output-1', variables=PRESELECT)
Model.historyOutputRequests['H-Output-1'].setValues(numIntervals=300, timeMarks=OFF)

# Initialize contact by eliminating the gap delta
Model.DisplacementBC(
    amplitude=UNSET, 
    createStepName='Make Contact', 
    distributionType=UNIFORM, 
    fieldName='', fixed=OFF, 
    localCsys=None, 
    name='BC-Periodic-1', 
    region=Region(referencePoints=(Model.rootAssembly.instances[RP1].referencePoints[1],)), 
    u1=0.0,
    u2=0.0,
    u3=0.0)

Model.DisplacementBC(
    amplitude=UNSET, 
    createStepName='Make Contact', 
    distributionType=UNIFORM, 
    fieldName='', fixed=OFF, 
    localCsys=None, 
    name='BC-Periodic-2', 
    region=Region(referencePoints=(Model.rootAssembly.instances[RP2].referencePoints[1],)),  
    u1=0,
    u2=gap,
    u3=0)

Model.DisplacementBC(
    amplitude=UNSET, 
    createStepName='Make Contact', 
    distributionType=UNIFORM, 
    fieldName='', fixed=OFF, 
    localCsys=None, 
    name='BC-Periodic-3', 
    region=Region(referencePoints=(Model.rootAssembly.instances[RP3].referencePoints[1],)),  
    u1=0,
    u2=0,
    u3=0)

# Deactivate the BC in the pressure step
Model.boundaryConditions['BC-Periodic-1'].deactivate('Apply Pressure')
Model.boundaryConditions['BC-Periodic-2'].deactivate('Apply Pressure')
Model.boundaryConditions['BC-Periodic-3'].deactivate('Apply Pressure')

# Apply pressure at the surfaces
Model.Pressure(
    amplitude=UNSET, 
    createStepName='Apply Pressure', 
    distributionType=UNIFORM, 
    field='', 
    magnitude=p, 
    name='PX', 
    region=Model.rootAssembly.instances['Layers'].surfaces['Surf-X'])

Model.Pressure(
    amplitude=UNSET, 
    createStepName='Apply Pressure', 
    distributionType=UNIFORM, 
    field='', 
    magnitude=p, 
    name='PY', 
    region=Model.rootAssembly.instances['Layers'].surfaces['Surf-Y'])


Model.Pressure(
    amplitude=UNSET, 
    createStepName='Apply Pressure', 
    distributionType=UNIFORM, 
    field='', 
    magnitude=p, 
    name='PZ', 
    region=Model.rootAssembly.instances['Layers'].surfaces['Surf-Z'])


# Make job
mdb.Job(
    atTime=None, 
    contactPrint=OFF, 
    description='', 
    echoPrint=OFF,
    explicitPrecision=SINGLE,  
    getMemoryFromAnalysis=True,
    historyPrint=OFF, 
    memory=100, 
    memoryUnits=PERCENTAGE, 
    model=ModelName, 
    modelPrint=OFF,
    multiprocessingMode=THREADS, 
    name=JobName, 
    nodalOutputPrecision=SINGLE,
    numCpus=ncpus, 
    numDomains=ncpus, 
    numGPUs=0, 
    queue=None, 
    scratch='', 
    type=ANALYSIS, 
    userSubroutine='', 
    waitHours=0, 
    waitMinutes=0)

# Save the CAE file
mdb.saveAs(pathName=Savename)

if SUBMIT:
    mdb.jobs[JobName].submit(consistencyChecking=OFF)
