__author__ = "N. Vasios"
__email__ = "vasios@g.harvard.edu"

'''
Numerical Analysis of Periodic Laminar and Fibrous media undergoing aJamming Transition
N. Vasios, B. Aktas, Y. Narang, R. Howe, K. Bertoldi

Abaqus Python script to post-process the results of each deformation simulation
'''

from abaqus import *
from abaqusConstants import *
import __main__

import section
import regionToolset
import displayGroupMdbToolset as dgm
import part
import material
import assembly
import step
import interaction
import load
import mesh
import optimization
import job
import sketch
import visualization
import xyPlot
import displayGroupOdbToolset as dgo
import connectorBehavior
import numpy as np



def ltuple_2_lnumpy(lttuple):

    '''
    When we extract data from the odb file, abaqus returns a list with
    the results for all the variables we requested. The list contains
    dictionaries and the data of the variables may be accessed as 
    .data

    For instance, consider that we request the stress and strain 
    from the odb as

    vars = session.xyDataListFromField(... ((STRESS),(STRAIN),)...)

    where now s = vars[0] and e = vars[1]

    the data for s and e are accessed as 

    sdata = s.data
    edata = e.data

    Now the variables sdata, edata are tuples of tuples, since abaqus returns
    the results as functions of time in the problem. The firsT ''column'' of
    the tuple is always time and the second is the variable we requested.

    We would like to eliminate time overall from the variables and also convert
    the tuples to numpy arrays. This function makes this exact conversion
    '''

    import numpy as np

    l = len(lttuple)

    lnumpy = [None for i in range(l)]
    for i in range(l):
        aux = lttuple[i].data

        lnumpy[i] = np.array([aux[j][1] for j in range(len(aux))])

    return lnumpy


# Name of odb file to be processed (name of the master file)
odbmaster = 'Fiber_3D_Square_fillet_0-'

# Extensio for each deformation case
names = ['u11','u12','u13','u21','u22','u23','u31','u32','u33']


# Loop over each file, for each deformation case
for i,name in enumerate(names):

    odbname = odbmaster+name+'.odb'
    o = session.openOdb(name=odbmaster+name+'.odb')

    odb = session.odbs[odbname]

    FD_tuple = session.XYDataFromHistory(
        name='ALLFD Whole Model-1', 
        odb=odb, 
        outputVariableName='Frictional dissipation: ALLFD for Whole Model', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    SE_tuple = session.XYDataFromHistory(
        name='ALLSE Whole Model-1', 
        odb=odb, 
        outputVariableName='Strain energy: ALLSE for Whole Model', 
        steps=('Deform-pos', 'Deform-neg', ),
        __linkedVpName__='Viewport: 1')

    U11_tuple = session.XYDataFromHistory(
        name='U1 PI: REFERENCE-POINT-1 N: 1 NSET REFERENCE-POINT-1-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U1 PI: REFERENCE-POINT-1 Node 1 in NSET REFERENCE-POINT-1', 
        steps=('Deform-pos', 'Deform-neg', ),
        __linkedVpName__='Viewport: 1')

    U12_tuple = session.XYDataFromHistory(
        name='U1 PI: REFERENCE-POINT-2 N: 1 NSET REFERENCE-POINT-2-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U1 PI: REFERENCE-POINT-2 Node 1 in NSET REFERENCE-POINT-2', 
        steps=('Deform-pos', 'Deform-neg', ),
         __linkedVpName__='Viewport: 1')

    U13_tuple = session.XYDataFromHistory(
        name='U1 PI: REFERENCE-POINT-3 N: 1 NSET REFERENCE-POINT-3-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U1 PI: REFERENCE-POINT-3 Node 1 in NSET REFERENCE-POINT-3', 
        steps=('Deform-pos', 'Deform-neg', ),
         __linkedVpName__='Viewport: 1')

    U21_tuple = session.XYDataFromHistory(
        name='U2 PI: REFERENCE-POINT-1 N: 1 NSET REFERENCE-POINT-1-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U2 PI: REFERENCE-POINT-1 Node 1 in NSET REFERENCE-POINT-1', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    U22_tuple = session.XYDataFromHistory(
        name='U2 PI: REFERENCE-POINT-2 N: 1 NSET REFERENCE-POINT-2-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U2 PI: REFERENCE-POINT-2 Node 1 in NSET REFERENCE-POINT-2', 
        steps=('Deform-pos', 'Deform-neg', ),
         __linkedVpName__='Viewport: 1')

    U23_tuple = session.XYDataFromHistory(
        name='U2 PI: REFERENCE-POINT-3 N: 1 NSET REFERENCE-POINT-3-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U2 PI: REFERENCE-POINT-3 Node 1 in NSET REFERENCE-POINT-3', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    U31_tuple = session.XYDataFromHistory(
        name='U3 PI: REFERENCE-POINT-1 N: 1 NSET REFERENCE-POINT-1-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U3 PI: REFERENCE-POINT-1 Node 1 in NSET REFERENCE-POINT-1', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    U32_tuple = session.XYDataFromHistory(
        name='U3 PI: REFERENCE-POINT-2 N: 1 NSET REFERENCE-POINT-2-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U3 PI: REFERENCE-POINT-2 Node 1 in NSET REFERENCE-POINT-2', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    U33_tuple = session.XYDataFromHistory(
        name='U3 PI: REFERENCE-POINT-3 N: 1 NSET REFERENCE-POINT-3-1', 
        odb=odb, 
        outputVariableName='Spatial displacement: U3 PI: REFERENCE-POINT-3 Node 1 in NSET REFERENCE-POINT-3', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF11_tuple = session.XYDataFromHistory(
        name='TF1 PI: REFERENCE-POINT-1 N: 1 NSET REFERENCE-POINT-1-1', 
        odb=odb, 
        outputVariableName='Total force: TF1 PI: REFERENCE-POINT-1 Node 1 in NSET REFERENCE-POINT-1', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF12_tuple = session.XYDataFromHistory(
        name='TF1 PI: REFERENCE-POINT-2 N: 1 NSET REFERENCE-POINT-2-1', 
        odb=odb, 
        outputVariableName='Total force: TF1 PI: REFERENCE-POINT-2 Node 1 in NSET REFERENCE-POINT-2', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF13_tuple = session.XYDataFromHistory(
        name='TF1 PI: REFERENCE-POINT-3 N: 1 NSET REFERENCE-POINT-3-1', 
        odb=odb, 
        outputVariableName='Total force: TF1 PI: REFERENCE-POINT-3 Node 1 in NSET REFERENCE-POINT-3', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF21_tuple = session.XYDataFromHistory(
        name='TF2 PI: REFERENCE-POINT-1 N: 1 NSET REFERENCE-POINT-1-1', 
        odb=odb, 
        outputVariableName='Total force: TF2 PI: REFERENCE-POINT-1 Node 1 in NSET REFERENCE-POINT-1', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF22_tuple = session.XYDataFromHistory(
        name='TF2 PI: REFERENCE-POINT-2 N: 1 NSET REFERENCE-POINT-2-1', 
        odb=odb, 
        outputVariableName='Total force: TF2 PI: REFERENCE-POINT-2 Node 1 in NSET REFERENCE-POINT-2', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF23_tuple = session.XYDataFromHistory(
        name='TF2 PI: REFERENCE-POINT-3 N: 1 NSET REFERENCE-POINT-3-1', 
        odb=odb, 
        outputVariableName='Total force: TF2 PI: REFERENCE-POINT-3 Node 1 in NSET REFERENCE-POINT-3', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    TF31_tuple = session.XYDataFromHistory(
        name='TF3 PI: REFERENCE-POINT-1 N: 1 NSET REFERENCE-POINT-1-1', 
        odb=odb, 
        outputVariableName='Total force: TF3 PI: REFERENCE-POINT-1 Node 1 in NSET REFERENCE-POINT-1', 
        steps=('Deform-pos', 'Deform-neg', ),
        __linkedVpName__='Viewport: 1')

    TF32_tuple = session.XYDataFromHistory(
        name='TF3 PI: REFERENCE-POINT-2 N: 1 NSET REFERENCE-POINT-2-1', 
        odb=odb, 
        outputVariableName='Total force: TF3 PI: REFERENCE-POINT-2 Node 1 in NSET REFERENCE-POINT-2', 
        steps=('Deform-pos', 'Deform-neg', ),
         __linkedVpName__='Viewport: 1')

    TF33_tuple = session.XYDataFromHistory(
        name='TF3 PI: REFERENCE-POINT-3 N: 1 NSET REFERENCE-POINT-3-1', 
        odb=odb, 
        outputVariableName='Total force: TF3 PI: REFERENCE-POINT-3 Node 1 in NSET REFERENCE-POINT-3', 
        steps=('Deform-pos', 'Deform-neg', ), 
        __linkedVpName__='Viewport: 1')

    # Convert to numpy arrays
    SEFD = ltuple_2_lnumpy([SE_tuple,FD_tuple])
    Us = ltuple_2_lnumpy([U11_tuple,U12_tuple,U13_tuple,U21_tuple,U22_tuple,U23_tuple,U31_tuple,U32_tuple,U33_tuple])
    TFs = ltuple_2_lnumpy([TF11_tuple,TF12_tuple,TF13_tuple,TF21_tuple,TF22_tuple,TF23_tuple,TF31_tuple,TF32_tuple,TF33_tuple])

    # Save as txt
    np.savetxt("{0}{1}.out".format(odbmaster,name),np.column_stack((Us[0],Us[1],Us[2],Us[3],Us[4],Us[5],Us[6],Us[7],Us[8],TFs[0],TFs[1],TFs[2],TFs[3],TFs[4],TFs[5],TFs[6],TFs[7],TFs[8],SEFD[0],SEFD[1])))




