__author__ = "N. Vasios"
__email__ = "vasios@g.harvard.edu"

'''
Numerical Analysis of Periodic Laminar and Fibrous media undergoing aJamming Transition
N. Vasios, B. Aktas, Y. Narang, R. Howe, K. Bertoldi

Abaqus python script for the restart analysis of RVEs
'''

from part import *
from material import *
from section import *
from assembly import *
from step import *
from interaction import *
from load import *
from mesh import *
from optimization import *
from job import *
from sketch import *
from visualization import *
from connectorBehavior import *
import numpy as np


execfile("utilities.py")

# Displacement Magnitude
u = 1e-1

# Reference point names
RP1 = 'Reference-Point-1'
RP2 = 'Reference-Point-2'
RP3 = 'Reference-Point-3'

# Deformation names
names = ['u11','u12','u13','u21','u22','u23','u31','u32','u33']

# Displacement Matrices for each case
Hmats = [np.array([[u,0,0],[0,0,0],[0,0,0]]),\
         np.array([[0,0,0],[u,0,0],[0,0,0]]),\
         np.array([[0,0,0],[0,0,0],[u,0,0]]),\
         np.array([[0,u,0],[0,0,0],[0,0,0]]),\
         np.array([[0,0,0],[0,u,0],[0,0,0]]),\
         np.array([[0,0,0],[0,0,0],[0,u,0]]),\
         np.array([[0,0,u],[0,0,0],[0,0,0]]),\
         np.array([[0,0,0],[0,0,u],[0,0,0]]),\
         np.array([[0,0,0],[0,0,0],[0,0,u]])]

# Model-name
ModelName = 'Model-1'

# Name of the origin - pressure job
JobName_Master = 'Fiber_3D_Square_fillet_0'

# Directories and cpus
basename = os.path.dirname(os.path.realpath('__file__'))  # Getting current directory
os.chdir(basename)  # Change to this directory
ncpus = 8 # Number of cpus
SUBMIT = True

# Run cases one by one
for i in range(len(names)):

	# Get name and displacememt matrix
    name = names[i]
    Hmat = Hmats[i]

    # Update Jobname
    JobName = JobName_Master + '-' + name

    # Open the Model Database
    mdb.openAuxMdb(pathName=basename+'\\'+JobName_Master+'.cae')

    # Copy the model
    mdb.copyAuxMdbModel(fromName='Model-1', toName='Model-1-Deform-{0}'.format(name))

    # Close the old model
    mdb.closeAuxMdb()

    # Make the Deform model
    Model_Deform = mdb.models['Model-1-Deform-{0}'.format(name)]

    # Indicate that this is a restart analysis
    Model_Deform.setValues(
        restartJob=JobName_Master, 
        restartStep='Apply Pressure')
        # restartIncrement=55)

    # Make the deformation step
    Model_Deform.StaticStep(
        timePeriod=1.0,
        initialInc=0.00001, 
        name='Deform-pos', 
        maxInc=0.01,
        minInc=1e-10,
        maxNumInc=10000,
        nlgeom=ON, 
        previous='Apply Pressure')

    # Define the BCs in general through the reference points
    # Velocity BC ensure that nodes that need to stay fixed, stay fixed by setting v=0
    Model_Deform.VelocityBC(
        amplitude=UNSET, 
        createStepName='Deform-pos', 
        distributionType=UNIFORM, 
        fieldName='', 
        localCsys=None, 
        name='BC-Periodic-1', 
        region=Region(referencePoints=(Model_Deform.rootAssembly.instances[RP1].referencePoints[1],)),
        v1=-Hmat[0,0], 
        v2=-Hmat[0,1], 
        v3=-Hmat[0,2])

    Model_Deform.VelocityBC(
        amplitude=UNSET, 
        createStepName='Deform-pos', 
        distributionType=UNIFORM, 
        fieldName='', 
        localCsys=None, 
        name='BC-Periodic-2', 
        region=Region(referencePoints=(Model_Deform.rootAssembly.instances[RP2].referencePoints[1],)),
        v1=-Hmat[1,0], 
        v2=-Hmat[1,1], 
        v3=-Hmat[1,2])

    Model_Deform.VelocityBC(
        amplitude=UNSET, 
        createStepName='Deform-pos', 
        distributionType=UNIFORM, 
        fieldName='', 
        localCsys=None, 
        name='BC-Periodic-3', 
        region=Region(referencePoints=(Model_Deform.rootAssembly.instances[RP3].referencePoints[1],)),
        v1=-Hmat[2,0], 
        v2=-Hmat[2,1], 
        v3=-Hmat[2,2])

    # Make a second step to deform in the opposite direction
    Model_Deform.StaticStep(
        timePeriod=2.0,
        initialInc=0.00001, 
        name='Deform-neg', 
        maxInc=0.01,
        minInc=1e-10,
        maxNumInc=10000,
        nlgeom=ON, 
        previous='Deform-pos')

    # Update the BCs for the second step
    Model_Deform.boundaryConditions['BC-Periodic-1'].setValuesInStep(
        stepName='Deform-neg',
        v1=Hmat[0,0], 
        v2=Hmat[0,1],
        v3=Hmat[0,2])

    Model_Deform.boundaryConditions['BC-Periodic-2'].setValuesInStep(
        stepName='Deform-neg',
        v1=Hmat[1,0], 
        v2=Hmat[1,1],
        v3=Hmat[1,2])

    Model_Deform.boundaryConditions['BC-Periodic-3'].setValuesInStep(
        stepName='Deform-neg',
        v1=Hmat[2,0], 
        v2=Hmat[2,1],
        v3=Hmat[2,2])

    # Make job
    mdb.Job(
        atTime=None, 
        contactPrint=OFF, 
        description='', 
        echoPrint=OFF,
        explicitPrecision=SINGLE,  
        getMemoryFromAnalysis=True,
        historyPrint=OFF, 
        memory=100, 
        memoryUnits=PERCENTAGE, 
        model=Model_Deform, 
        modelPrint=OFF,
        multiprocessingMode=THREADS, 
        name=JobName, 
        nodalOutputPrecision=SINGLE,
        numCpus=ncpus, 
        numDomains=ncpus, 
        numGPUs=0, 
        queue=None, 
        scratch='', 
        type=RESTART, 
        userSubroutine='', 
        waitHours=0, 
        waitMinutes=0)

    # Submit the job and wait for completion before proceeding with the next one
    mdb.jobs[JobName].submit(consistencyChecking=OFF)
    mdb.jobs[JobName].waitForCompletion()

