__author__ = "N. Vasios"
__email__ = "vasios@g.harvard.edu"

'''
Numerical Analysis of Periodic Laminar and Fibrous media undergoing aJamming Transition
N. Vasios, B. Aktas, Y. Narang, R. Howe, K. Bertoldi

Function to enforce periodic boundary conditions in 3d bodies
'''


def PBC3D(Model,Set_Slave,Set_Master,LatticeVec,dim,cp=0, RP1=None, RP2=None, RP3=None):

    '''
    A function to enforce periodic boundary conditions in the form of equations

    Input arguments: 

        Model: (Object) The model to work on
        Set_Slave: (String) The name of the set with the slave nodes
        Set_Master: (String) The name of the set with the master nodes
        LatticeVec: (numpy matrix 3x3) The 3 lattice vectors in matrix form
        dim: (Int) Dimension along witch the PBC will be enforced
                     1 - along X
                     2 - along Y
                     3 - along Z
                    -1 - edges - all coupled with the master edge
                    -2 - corners - all coupled with the master corner
        cp: (Int) Number of constraint pairs already defined (To make names unique)
        RP1: (String) Name of the first reference point
        RP2: (String) Name of the second reference point
        RP3: (String) Name of the third reference point
    '''

    from part import THREE_D, DEFORMABLE_BODY 

    # If no reference point information was passed in, define them here
    if RP1 is None:
        RP1 = 'Reference-Point-1'
    if RP2 is None:
        RP2 = 'Reference-Point-2'
    if RP3 is None:
        RP3 = 'Reference-Point-3'

    tol = 1e-5

    if cp==0:

        # If this is constraint number 0 then introduce the reference points

        Model.Part(dimensionality=THREE_D, name=RP1, type=DEFORMABLE_BODY)
        Model.Part(dimensionality=THREE_D, name=RP2, type=DEFORMABLE_BODY)
        Model.Part(dimensionality=THREE_D, name=RP3, type=DEFORMABLE_BODY)

        Model.parts[RP1].ReferencePoint(point=(0.0, 0.0, 0.0))
        Model.parts[RP2].ReferencePoint(point=(0.0, 0.0, 0.0))
        Model.parts[RP3].ReferencePoint(point=(0.0, 0.0, 0.0))


        Model.rootAssembly.Instance(
            dependent=ON, 
            name=RP1, 
            part=Model.parts[RP1]
            )

        Model.rootAssembly.Instance(
            dependent=ON, 
            name=RP2, 
            part=Model.parts[RP2]
            )

        Model.rootAssembly.Instance(
            dependent=ON, 
            name=RP3, 
            part=Model.parts[RP3]
            )

        #Create set of reference points
        Model.rootAssembly.Set(
            name=RP1,
            referencePoints=(Model.rootAssembly.instances[RP1].referencePoints[1],)
            )

        Model.rootAssembly.Set(
            name=RP2, 
            referencePoints=(Model.rootAssembly.instances[RP2].referencePoints[1],)
            )

        Model.rootAssembly.Set(
            name=RP3, 
            referencePoints=(Model.rootAssembly.instances[RP3].referencePoints[1],)
            )

    if dim==-2:
        # Handling Corner nodes
        Nodes_slave  = Model.rootAssembly.sets[Set_Slave].nodes
        Nodes_master = Model.rootAssembly.sets[Set_Master].nodes
        mnode = Nodes_master[0]

        # Constrain each slave corver node to the master corner node
        for snode in Nodes_slave:

            dx = snode.coordinates[0] - mnode.coordinates[0]
            dy = snode.coordinates[1] - mnode.coordinates[1]
            dz = snode.coordinates[2] - mnode.coordinates[2]

            cp+=1

            cpname = str(cp)

            setname1 = Set_Master+'_N{0}P{1}'.format('1',cpname)
            setname2 = Set_Master+'_N{0}P{1}'.format('2',cpname)

            constraint_name_1 = Set_Master+'_P{0}E{1}'.format(cpname,'X')
            constraint_name_2 = Set_Master+'_P{0}E{1}'.format(cpname,'Y')
            constraint_name_3 = Set_Master+'_P{0}E{1}'.format(cpname,'Z')


            Model.rootAssembly.SetFromNodeLabels(
                name=setname1, 
                nodeLabels=((snode.instanceName, (snode.label,)),))                 

            Model.rootAssembly.SetFromNodeLabels(
                name=setname2, 
                nodeLabels=((mnode.instanceName, (mnode.label,)),)) 

            Model.Equation(
                name=constraint_name_1, 
                terms=(
                    (1.0, setname1, 1), 
                    (-1.0, setname2, 1), 
                    (-dx, RP1, 1), 
                    (-dy, RP1, 2), 
                    (-dz, RP1, 3)
                    )
                )


            Model.Equation(
                name=constraint_name_2,
                terms=(
                    ( 1.0,setname1, 2),
                    (-1.0,setname2, 2),
                    ( -dx,RP2, 1),
                    ( -dy,RP2, 2),
                    ( -dz,RP2, 3)
                    )
                )

            Model.Equation(
                name=constraint_name_3,
                terms=(
                    ( 1.0,setname1, 3),
                    (-1.0,setname2, 3),
                    ( -dx,RP3, 1),
                    ( -dy,RP3, 2),
                    ( -dz,RP3, 3)
                    )
                )

        return (RP1,RP2,RP3,cp)


    if dim==-1:

        # Handling Edge nodes
        Nodes_slave  = Model.rootAssembly.sets[Set_Slave].nodes
        Nodes_master = Model.rootAssembly.sets[Set_Master].nodes

        # Iterate ove each node in the slave set
        for mnode in Nodes_master:

            k = 0

            # Constrain each slave edge node to the master corner node
            for snode in Nodes_slave:

                dx = snode.coordinates[0] - mnode.coordinates[0]
                dy = snode.coordinates[1] - mnode.coordinates[1]
                dz = snode.coordinates[2] - mnode.coordinates[2]

                dv = np.array([dx,dy,dz])

                LV1
                LV2
                LV3
                LV1+LV2
                LV2+LV3
                LV1+LV3
                LV1+LV2+LV3

                if np.linalg.norm(LV1+dv)<=tol or np.linalg.norm(LV2+dv)<=tol or np.linalg.norm(LV3+dv)<=tol or np.linalg.norm(LV1+LV2+dv)<=tol or np.linalg.norm(LV2+LV3+dv)<=tol or np.linalg.norm(LV1+LV3+dv)<=tol or np.linalg.norm(LV1+LV2+LV3+dv)<=tol:

                    cp+=1

                    cpname = str(cp)

                    setname1 = Set_Master+'_N{0}P{1}'.format('1',cpname)
                    setname2 = Set_Master+'_N{0}P{1}'.format('2',cpname)

                    constraint_name_1 = Set_Master+'_P{0}E{1}'.format(cpname,'X')
                    constraint_name_2 = Set_Master+'_P{0}E{1}'.format(cpname,'Y')
                    constraint_name_3 = Set_Master+'_P{0}E{1}'.format(cpname,'Z')

                    Model.rootAssembly.SetFromNodeLabels(
                        name=setname1, 
                        nodeLabels=((snode.instanceName, (snode.label,)),))                    

                    Model.rootAssembly.SetFromNodeLabels(
                        name=setname2, 
                        nodeLabels=((mnode.instanceName, (mnode.label,)),))

                    Model.Equation(
                        name=constraint_name_1, 
                        terms=(
                            (1.0, setname1, 1), 
                            (-1.0, setname2, 1), 
                            (-dx, RP1, 1), 
                            (-dy, RP1, 2), 
                            (-dz, RP1, 3)
                            )
                        )


                    Model.Equation(
                        name=constraint_name_2,
                        terms=(
                            ( 1.0,setname1, 2),
                            (-1.0,setname2, 2),
                            ( -dx,RP2, 1),
                            ( -dy,RP2, 2),
                            ( -dz,RP2, 3)
                            )
                        )

                    Model.Equation(
                        name=constraint_name_3,
                        terms=(
                            ( 1.0,setname1, 3),
                            (-1.0,setname2, 3),
                            ( -dx,RP3, 1),
                            ( -dy,RP3, 2),
                            ( -dz,RP3, 3)
                            )
                        )

                    k = k + 1

                if k == 3:
                    break

                        
        return (RP1,RP2,RP3,cp)

    else:

        # Handling Other nodes

        Nodes_slave  = Model.rootAssembly.sets[Set_Slave].nodes
        Nodes_master = Model.rootAssembly.sets[Set_Master].nodes

        # Loop over each node in the master set
        for snode in Nodes_slave:

            # Constrain each slave node to the corresponding master
            for mnode in Nodes_master:

                dx = snode.coordinates[0] - mnode.coordinates[0]
                dy = snode.coordinates[1] - mnode.coordinates[1]
                dz = snode.coordinates[2] - mnode.coordinates[2]

                if abs(LatticeVec[dim-1,0]+dx) + abs(LatticeVec[dim-1,1]+dy) + abs(LatticeVec[dim-1,2]+dz) <=tol:

                    cp+=1

                    cpname = str(cp)

                    setname1 = Set_Master+'_N{0}P{1}'.format('1',cpname)
                    setname2 = Set_Master+'_N{0}P{1}'.format('2',cpname)

                    constraint_name_1 = Set_Master+'_P{0}E{1}'.format(cpname,'X')
                    constraint_name_2 = Set_Master+'_P{0}E{1}'.format(cpname,'Y')
                    constraint_name_3 = Set_Master+'_P{0}E{1}'.format(cpname,'Z')


                    Model.rootAssembly.SetFromNodeLabels(
                        name=setname1, 
                        nodeLabels=((snode.instanceName, (snode.label,)),))                    

                    Model.rootAssembly.SetFromNodeLabels(
                        name=setname2, 
                        nodeLabels=((mnode.instanceName, (mnode.label,)),))

                    Model.Equation(
                        name=constraint_name_1, 
                        terms=(
                            (1.0, setname1, 1), 
                            (-1.0, setname2, 1), 
                            (-dx, RP1, 1), 
                            (-dy, RP1, 2), 
                            (-dz, RP1, 3)
                            )
                        )


                    Model.Equation(
                        name=constraint_name_2,
                        terms=(
                            ( 1.0,setname1, 2),
                            (-1.0,setname2, 2),
                            ( -dx,RP2, 1),
                            ( -dy,RP2, 2),
                            ( -dz,RP2, 3)
                            )
                        )

                    Model.Equation(
                        name=constraint_name_3,
                        terms=(
                            ( 1.0,setname1, 3),
                            (-1.0,setname2, 3),
                            ( -dx,RP3, 1),
                            ( -dy,RP3, 2),
                            ( -dz,RP3, 3)
                            )
                        )

                    break         

        return (RP1,RP2,RP3,cp)